/* ATest.cpp
called by Aardvark to test stuff
*/

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// incorporate it inside of an executable or other machine language component
// without restriction for personal or commercial use.The Software may not be
// redistributed as source code or in any recognizable human readable form
// in any form whatsoever for any use whatsoever.
// 
// The Software is the property of Michle Delsol(France), copyright 2023 USA and
// international. For any questions concerning use of the software contact
// Michle Delsol at CPParduino@md-dsl.fr.
// 
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

// Scenario:
// A set of sensor variables is transformed into a byte array so that 
// it might be transmitted as a sequence of bytes, via a serial port or via radio.
// Conversion from float arrays to int arrays and vice versa are handled by
// the BFFDC class.
// There are two distinct situations:
//  1 - Transform a set of sensor float values into corresponding uint16_t values and
//  compress the uint16_t values into uint8_t values when value is 254 or less.
//  2 - Transform set of compressed uint8_t values into uint16_t values which 
//  are then converted into floats.

#include "Globals.h"

//#define _DEBUG_ATEST

// Array of sensor values to convert to bytes for data transmission
// Test specific data for one hive
float sensorValue[]{
	7.35, // power8V
	4.92, // power5V
	34.1, // temp
	71, // humid
	3.5, // weightBody
	0.2 // weightSuper
};

char* sensorLabels[] = { "8V", "5V", "temp", "humid", "weightBody", "weightSuper" };
char* GetSensorLabel(EnumSensors _sensorID) { return sensorLabels[(uint8_t)_sensorID]; };

// Possible values
// item        range    decimals  divide 
// bits in control values   2          1      
// 8V        5.0 to 8.5     1      false     // 6.87V, 8.35V... no need for 2nd decimal
// 5V        4.0 to 5.5     1      false     // 4.34V, 4.46V... no need for 2nd decimal
// temp       20 to  60     0      false     // 32.5 - no need for decimals
// humid      30 to 100     1       true     // 30%, 40%... sufficient
// weights - 5 volts equal 100 kgs, two decimals, 500 values, allows a 0.2kg resolution
// body     0.05 to 5.00    2      false     //
// super    0.05 to 5.00    2      false     // 

// Min/max to reduce mem requirements
float sensorRange[][2] = { // 
	{5.0, 9.0}, // Voltage8V
   {4.0, 6.0}, // Voltage5V
   {-20.0, 60.0}, // Temp
   {30, 100.0}, // Humid - in tens thus 10 is really 100
   {0.0, 5.0}, // WeightBody
   {0.0, 5.0} // WeightSuper
};


// Bit masks to get control variables values
uint8_t sensorControlValues[] {
  B01000000, // 8V: 1 decimal, , divide No
  B01000000, // 5V: 1 decimal, , divide No
  B00000000, // Temp: 0 decimal, divide No
  B01100000, // Humid: 1 decimal, divide Yes
  B10000000, // WeightBody: 2 decimal, divide No
  B10000000 // WeightSuper: 2 decimal, divide No
};

// Destination arrays when converting from floats/uint8_t or vice versa
uint8_t* convertedFloatsArray{ 0 };
uint8_t* converted_uint8_tArray{ 0 };


void ATest() {
	bool success{ false };
	uint8_t size{ 0 };
	bool allGood{ true };
	uint8_t noSensors = (uint8_t)EnumSensors::end;
	float destinationFloats[noSensors];
	Serial.println(F("Starting - print values (sensorValue array)."));
	for (int i = 0; i < noSensors; i++) {
		Serial.print(sensorLabels[i]); Serial.print(F(" - "));
		Serial.println(sensorValue[i]);
	}

	BFFDC* bffdc[noSensors];
	// Array values once the conversion process has finished
	uint16_t rawUint16data[noSensors];
	// Most variables, once compacted, require one byte only; some require two
	uint8_t bytesRequiredIndividualArray[noSensors]; // Bytes required for individual variables
	// Most variables, once compacted, require one byte only; some require two
	uint8_t bytesRequiredTotal{ 0 }; // Total bytes required for data transfer
	// Array to be used in data transfer, oversized - bytesRequiredTotal will be used
	uint8_t compactedUint8Array[(noSensors)*2]; 

	// 1 - Transform a set of sensor float values into corresponding  uint16_t values and
	// compress the uint16_t values into uint8_t values when value is 254 or less.
	for (int i = 0; i < noSensors; i++) {
		// Create BFFDC instances and get the number of bytes required for the assembled 
		// values into a byte array
		//Serial.print(F("Creating BFFDC instance for ")); Serial.println(sensorLabels[i]);
		bffdc[i] = new BFFDC((bool&)allGood, sensorValue[i], (float*)sensorRange[i], sensorControlValues[i], sensorLabels[i]);
		// If constructor detects error sets allGood to false
		if (allGood == false) {
			Serial.println(F("BFFDC returned false - Aborting"));
			return;
		} // if (allGood == false)
		//bffdc[i]->PrintValuesInControlVariables();
		uint8_t bytesRequired = bffdc[i]->BytesRequired();
		bytesRequiredIndividualArray[i] = bytesRequired;
		bytesRequiredTotal += bytesRequired;

		//Serial.print(F("Starting conversion on ")); Serial.print(sensorLabels[i]);
		//Serial.print(F(" - ")); Serial.println(sensorValue[i]);
		rawUint16data[i] = bffdc[i]->TransformFloatIntoInt16((bool&)allGood);
		//Serial.print(F("Conversion result ")); Serial.println(rawUint16data[i]);
		delete bffdc[i];
	} // for (int i = 0; i < noSensors; i++)

#ifdef _DEBUG_ATEST
	Serial.print(F("Total bytes required = ")); Serial.println(bytesRequiredTotal);
	for (int i = 0; i < noSensors; i++) {
		Serial.print(sensorLabels[i]); 
		Serial.print(F(": bytes required ")); Serial.print(bytesRequiredIndividualArray[i]);
		Serial.print(F(" - value ")); Serial.print(sensorValue[i]);
		Serial.print(F(" - raw uint16 result ")); Serial.print(rawUint16data[i]);
		Serial.println();
	}
#endif //_DEBUG_ATEST
	UAC uac = UAC(noSensors, bytesRequiredTotal, rawUint16data, bytesRequiredIndividualArray, compactedUint8Array);
	uac.ProcessRawToCompact();
#ifdef _DEBUG_ATEST
	Serial.println(F("\nTransfer results from raw uint16 array to uint8 array."));
	Serial.print(F("Float data reduced from "));
	Serial.print(noSensors * sizeof(float));
	Serial.print(F(" to "));
	Serial.print(bytesRequiredTotal);
	Serial.println(F(" located in compactedUint8Array. Values are:"));
	for (int i = 0; i < bytesRequiredTotal; i++) {
		Serial.print(F("i = "));
		Serial.print(i);
		Serial.print(F(" - "));
		Serial.println(compactedUint8Array[i]);
	} // for (int i = 0; i < noSensors; i++)
#endif //_DEBUG_ATEST
	// This finishes phase 1.


	// 2 - Transform set of compressed uint8_t values into uint16_t values which 
	// are then converted into floats.
	uac.ProcessCompactToRaw();
#ifdef _DEBUG_ATEST
	Serial.println(F("\nTransfer results from final uint8 array to float array."));
	Serial.println(F("Values in compactedUint16Array:"));
	for (int i = 0; i < noSensors; i++) {
		Serial.print(F("i = "));
		Serial.print(i);
		Serial.print(F(" - "));
		Serial.println(rawUint16data[i]);
	} // for (int i = 0; i < noSensors; i++)
#endif //_DEBUG_ATEST

	for (int i = 0; i < noSensors; i++) {
		bffdc[i] = new BFFDC((bool&)allGood, rawUint16data[i], (float*)sensorRange[i], sensorControlValues[i], sensorLabels[i]);
		destinationFloats[i] = bffdc[i]->TransformInt16IntoFloat();
		delete bffdc[i];
	} // for (int i = 0; i < noSensors; i++)
	Serial.println(F("Finished - print results (destinationFloats array)."));
	for (int i = 0; i < noSensors; i++) {
		Serial.print(sensorLabels[i]); Serial.print(F(" - "));
		Serial.println(destinationFloats[i]);
	} // for (int i = 0; i < noSensors; i++)

} // ATest()

// End ATest.cpp
