// Begin BFFDC.cpp

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// incorporate it inside of an executable or other machine language component
// without restriction for personal or commercial use.The Software may not be
// redistributed as source code or in any recognizable human readable form
// in any form whatsoever for any use whatsoever.
// 
// The Software is the property of Michle Delsol(France), copyright 2023 USA and
// international. For any questions concerning use of the software contact
// Michle Delsol at CPParduino@md-dsl.fr.
// 
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include "Globals.h"

#define _DEBUG_BFFDC_ON
#ifdef _DEBUG_BFFDC_ON

// UAC defines
//#define _DEBUG_UAC_UAC
//#define _DEBUG_UAC_PROCESS_RAW_TO_COMPACT
//#define _DEBUG_UAC_PROCESS_COMPACT_TO_RAW

// BFFDC defines
//#define _DEBUG_BFFDC_CONSTRUCTOR
//#define _DEBUG_BFFDC_DESTRUCTOR
//#define _DEBUG_BFFDC_BYTES_REQUIRED
//#define _DEBUG_BFFDC_GET_BFFDC_VALUE
//#define _DEBUG_BFFDC_GET_NUMBER_OF_DECIMALS
//#define _DEBUG_BFFDC_GET_NEGATIVE_OK
//#define _DEBUG_BFFDC_CONVERT_FLOAT_TO_UINT32
//#define _DEBUG_BFFDC_CONVERT_UINT32_TO_FLOAT
//#define _DEBUG_BFFDC_TRANSFER_FLOAT_TO_INT16_ARRAY
//#define _DEBUG_BFFDC_TRANSFER_INT16_TO_FLOAT
//#define _DEBUG_BFFDC_GET_CONTROL_VARIABLES_ARRAY_SIZE
//#define _DEBUG_BFFDC_GET_ALL_CONTROL_PARAMETERS
//#define _DEBUG_BFFDC_GET_OFFSET_SENSOR_IN_GLOBAL_ARRAY
//#define _DEBUG_BFFDC_GET_TOTAL_GLOBAL_ARRAY_SIZE
//#define _DEBUG_BFFDC_EXTRACT_VALUE_FROM_BITMAP

#endif

//*************************************************************************************
// The UAC class is a support class to transfer uncompacted rawUint16data array
// and the compactedUint8Array.
// Transfer is straight uint16 to uint8 assignment for data items which require
// one byte (_bytesRequiredIndividualArray).
// Transfer is separated as two bytes 254 based LSB/MSB (least significant, most significant) bytes
// for data items which require two bytes
// The work is performed inside the constructor, two methods enables getting results.

UAC::UAC(uint8_t _noItems, 
	uint8_t _bytesRequiredTotal, uint16_t* _rawUint16data,
	uint8_t* _bytesRequiredIndividualArray, uint8_t* _compactedUint8Array) :
	noItems(_noItems), 
	bytesRequiredTotal(_bytesRequiredTotal), rawUint16data(_rawUint16data),
	bytesRequiredIndividualArray(_bytesRequiredIndividualArray), compactedUint8Array(_compactedUint8Array) {
#ifdef _DEBUG_UAC_UAC
	PRINT_FUNCTION_ENTER_NL
#endif // _DEBUG_UAC_UAC
	// TODO Verify number of items
#ifdef _DEBUG_UAC_UAC
	PRINT_VALUE_EQUAL_DESC_NL("UAC - processing byte arrays on  ", noItems);
#endif // _DEBUG_UAC_UAC

#ifdef _DEBUG_UAC_UAC
	PRINT_FUNCTION_EXIT_NL
#endif // _DEBUG_UAC_UAC
} // UAC::UAC

void UAC::ProcessRawToCompact() {
#ifdef _DEBUG_UAC_PROCESS_RAW_TO_COMPACT
	PRINT_FUNCTION_EXIT_NL;
#endif // _DEBUG_UAC_PROCESS_RAW_TO_COMPACT
	uint8_t bytesRequired{ 0 };
	uint8_t offsetIntoUint8Array{ 0 };
	for (int i = 0; i < noItems; i++) {
		bytesRequired = bytesRequiredIndividualArray[i];
#ifdef _DEBUG_UAC_PROCESS_RAW_TO_COMPACT
		PRINT_VALUE_EQUAL_TAB_NL(bytesRequired);
#endif // _DEBUG_UAC_PROCESS_RAW_TO_COMPACT
		if (bytesRequired == 1) {
			compactedUint8Array[offsetIntoUint8Array] = rawUint16data[i];
#ifdef _DEBUG_UAC_PROCESS_RAW_TO_COMPACT
			PRINT_VALUE_EQUAL_TAB_NL(compactedUint8Array[offsetIntoUint8Array]);
#endif // _DEBUG_UAC_PROCESS_RAW_TO_COMPACT
		}
		else {
			uint8_t LSB = rawUint16data[i] % 254;
			uint8_t MSB = rawUint16data[i] / 254;
			compactedUint8Array[offsetIntoUint8Array] = LSB;
			compactedUint8Array[offsetIntoUint8Array + 1] = MSB;
#ifdef _DEBUG_UAC_PROCESS_RAW_TO_COMPACT
			PRINT_VALUE_EQUAL_TAB_NL(compactedUint8Array[offsetIntoUint8Array]);
			PRINT_VALUE_EQUAL_TAB_NL(compactedUint8Array[offsetIntoUint8Array + 1]);
#endif // _DEBUG_UAC_PROCESS_RAW_TO_COMPACT
		}
		offsetIntoUint8Array += bytesRequired;
	}
#ifdef _DEBUG_UAC_PROCESS_COMPACT_TO_RAW
	PRINT_FUNCTION_EXIT_NL;
#endif // _DEBUG_UAC_PROCESS_RAW_TO_COMPACT
	return;
} // UAC::ProcessCompactToRaw

void UAC::ProcessCompactToRaw() {
#ifdef _DEBUG_UAC_PROCESS_COMPACT_TO_RAW
	PRINT_FUNCTION_ENTER_NL;
#endif // _DEBUG_UAC_PROCESS_COMPACT_TO_RAW
	uint8_t bytesRequired{ 0 };
	uint8_t offsetIntoUint8Array{ 0 };
	for (int i = 0; i < noItems; i++) {
		bytesRequired = bytesRequiredIndividualArray[i];
#ifdef _DEBUG_UAC_PROCESS_RAW_TO_COMPACT
		PRINT_VALUE_EQUAL_TAB_NL(bytesRequired);
#endif // _DEBUG_UAC_PROCESS_RAW_TO_COMPACT
		if (bytesRequired == 1) {
			rawUint16data[i] = compactedUint8Array[offsetIntoUint8Array];
		}
		else {
			uint8_t LSB = compactedUint8Array[offsetIntoUint8Array];
			uint8_t MSB = compactedUint8Array[offsetIntoUint8Array + 1];
			rawUint16data[i] = LSB + MSB * 254;
		}
#ifdef _DEBUG_UAC_PROCESS_RAW_TO_COMPACT
		PRINT_VALUE_EQUAL_TAB_NL(destinationFloatsArray[i]);
#endif // _DEBUG_UAC_PROCESS_RAW_TO_COMPACT
		offsetIntoUint8Array += bytesRequired;
	}

#ifdef _DEBUG_UAC_PROCESS_COMPACT_TO_RAW
	PRINT_FUNCTION_EXIT_NL;
#endif // _DEBUG_UAC_PROCESS_COMPACT_TO_RAW
	return;
} // UAC::ProcessCompactToRaw

//*************************************************************************************

char* BFFDC::controlLabels[] = {
  "decimals",
  "Divide",
  "Spare"
};

uint8_t BFFDC::controlVariableMasks[]{ 	// Control variables bit masks 
	B11000000, /* decimals */
	B00100000, /* divide */
	B00011111  /* Spare */
};

uint8_t BFFDC::controlVariableMaskOffset[]{ 	// Control variables bit masks 
	6, /* decimals */
	5, /* divide */
	0  /* Spare */
};

uint8_t BFFDC::ExtractValueFromBitmap(EControlVariables _varID) {
	// Extracts control operator from the control variables bitmap located in a uint8_t byte
	// Phase ONE is to get the mask from the controlVariableBitMasks array for the desired variable
	// Phase TWO is to apply the mask and shift on the variable's controVariables to get
	// the application dependent control variable
#ifdef _DEBUG_BFFDC_EXTRACT_VALUE_FROM_BITMAP
	PRINT_FUNCTION_ENTER_NL;
	Serial.print(F("Getting value for ")); Serial.print((char*)floatValueName);
	Serial.print(F(" - ")); Serial.println((char*)controlLabels[(uint8_t)_varID]);
	//PRINT_VALUE_EQUAL_TAB_BIN_NL(_sensorControlValues);
#endif // _DEBUG_BFFDC_EXTRACT_VALUE_FROM_BITMAP

	// Phase ONE
	uint8_t extractionMask = controlVariableMasks[(uint8_t)_varID];
	uint8_t controlVarOffset = controlVariableMaskOffset[(uint8_t)_varID];
#ifdef _DEBUG_BFFDC_EXTRACT_VALUE_FROM_BITMAP
	PRINT_VALUE_EQUAL_TAB_BIN_NL(extractionMask);
	PRINT_VALUE_EQUAL_TAB_NL(controlVarOffset);
#endif // _DEBUG_BFFDC_EXTRACT_VALUE_FROM_BITMAP

	// Phase TWO
uint8_t valueToGet{ 0 };
valueToGet = controlVariables & extractionMask;
valueToGet >>= controlVarOffset;
#ifdef _DEBUG_BFFDC_EXTRACT_VALUE_FROM_BITMAP
PRINT_VALUE_EQUAL_TAB_NL(valueToGet);
PRINT_FUNCTION_EXIT_NL;
#endif // _DEBUG_BFFDC_EXTRACT_VALUE_FROM_BITMAP
return valueToGet;
} // ExtractValueFromBitmap

void BFFDC::PrintValuesInControlVariables() {
	//PRINT_VALUE_EQUAL_DESC_NL("Getting value for ", sensorLabels[(uint8_t)_sensorID]);
	//uint8_t individualControlValues = sensorControlValues[(uint8_t)_sensorID];
	PRINT_VALUE_EQUAL_DESC_NL("Printing parameters for ", floatValueName);
	for (int i = 0; i < (int)EControlVariables::end; i++) {
		uint8_t valueToGet = ExtractValueFromBitmap((EControlVariables)i);
		PRINT_VALUE_EQUAL_TAB(controlLabels[i]);
		PRINT_VALUE_EQUAL_TAB_NL(valueToGet);
	}
} // PrintValuesInControlVariables

// Constructor gets float value, generates uint165_t value
BFFDC::BFFDC(bool& _errorCondition, float _floatValue, float _minMaxValues[], uint8_t _controlVariables, char* _floatValueName)
	: floatValue(_floatValue), minValue(_minMaxValues[0]), maxValue(_minMaxValues[1]), controlVariables(_controlVariables), floatValueName(_floatValueName) {
	// PROGRAM DOC START BFFDC()
	// ERROR CHECKING
	// Error checks total : 0
	// None
	// PARAMETERS
	// Parameters total : 2
	// uint8_t** _controlVariables - table of operators to format data 
	// uint8_t** _minMaxValues - minimum/maximum values for each variable
	// RETURNS
	// Returns bool - success/failure
	// PROGRAM DOC END BFFDC()
	// PROGRAM FLOW START calling function: BFFDC()

#ifdef _DEBUG_BFFDC_CONSTRUCTOR
	PRINT_FUNCTION_ENTER_NL;
	Serial.print(F("Handling item ")); Serial.println((char*)floatValueName);
#endif

	_errorCondition = true;

	// Check that value is within range
	decimals = ExtractValueFromBitmap(EControlVariables::decimals);
	divide = (bool)ExtractValueFromBitmap(EControlVariables::divide);

	if (floatValue < minValue || floatValue > maxValue) {
#ifdef _DEBUG_BFFDC_CONSTRUCTOR
		PROGRAM_ERROR_TEXT("BFFDC ERROR - floatValue out of range relative to min or Max values ");
		PRINT_VALUE_EQUAL_TAB_NL(floatValue);
		PRINT_VALUE_EQUAL_TAB_NL(minValue);
		PRINT_VALUE_EQUAL_TAB_NL(maxValue);
#endif
		_errorCondition = false;
	}

	if ((uint8_t)decimals < 0 || (uint8_t)decimals > 2) {
#ifdef _DEBUG_BFFDC_CONSTRUCTOR
		PROGRAM_ERROR_TEXT("BFFDC ERROR - decimals should lie between 0 and 2 ");
		PRINT_VALUE_EQUAL_TAB_NL(decimals);
#endif
		_errorCondition = false;
	}

#ifdef _DEBUG_BFFDC_CONSTRUCTOR
	PRINT_FUNCTION_EXIT_NL
#endif
		return;
	// PROGRAM FLOW END calling function: BFFDC()
} // BFFDC() constructor

// Constructor gets uint16_t value, generates float value
BFFDC::BFFDC(bool& _errorCondition, uint16_t _uint16Value, float _minMaxValues[], uint8_t _controlVariables, char* _floatValueName)
	: uint16Value(_uint16Value), minValue(_minMaxValues[0]), maxValue(_minMaxValues[1]), controlVariables(_controlVariables), floatValueName(_floatValueName) {
	// PROGRAM DOC START BFFDC()
	// ERROR CHECKING
	// Error checks total : 0
	// None
	// PARAMETERS
	// Parameters total : 2
	// uint8_t** _controlVariables - table of operators to format data 
	// uint8_t** _minMaxValues - minimum/maximum values for each variable
	// RETURNS
	// Returns bool - success/failure
	// PROGRAM DOC END BFFDC()
	// PROGRAM FLOW START calling function: BFFDC()

#ifdef _DEBUG_BFFDC_CONSTRUCTOR
	PRINT_FUNCTION_ENTER_NL;
	Serial.print(F("Handling item ")); Serial.println((char*)floatValueName);
#endif

	_errorCondition = true;

	// Check that value is within range
	decimals = ExtractValueFromBitmap(EControlVariables::decimals);
	divide = (bool)ExtractValueFromBitmap(EControlVariables::divide);

	if (floatValue < minValue || floatValue > maxValue) {
#ifdef _DEBUG_BFFDC_CONSTRUCTOR
		PROGRAM_ERROR_TEXT("BFFDC ERROR - floatValue out of range relative to min or Max values ");
		PRINT_VALUE_EQUAL_TAB_NL(floatValue);
		PRINT_VALUE_EQUAL_TAB_NL(minValue);
		PRINT_VALUE_EQUAL_TAB_NL(maxValue);
#endif
		_errorCondition = false;
	}

	if ((uint8_t)decimals < 0 || (uint8_t)decimals > 2) {
#ifdef _DEBUG_BFFDC_CONSTRUCTOR
		PROGRAM_ERROR_TEXT("BFFDC ERROR - decimals should lie between 0 and 2 ");
		PRINT_VALUE_EQUAL_TAB_NL(decimals);
#endif
		_errorCondition = false;
	}

#ifdef _DEBUG_BFFDC_CONSTRUCTOR
	PRINT_FUNCTION_EXIT_NL
#endif
		return;
	// PROGRAM FLOW END calling function: BFFDC()
} // BFFDC() constructor

BFFDC::~BFFDC() {

} // BFFDC() destructor

uint8_t BFFDC::BytesRequired() {
	// PROGRAM DOC START CalculateBytesRequiredFloat()
	// Determines the byte number required based on the max/min max range differential and 
	// number of decimals
	// ERROR CHECKING
	// Error checks total : 0
	// PARAMETERS
	// Parameters total : 0
	// RETURNS
	// Returns bytesRequired 2 if delta between max/min multiplied by decimals > 254.
	// PROGRAM DOC END BytesRequired()

#ifdef _DEBUG_BFFDC_BYTES_REQUIRED
	PRINT_FUNCTION_ENTER_NL;
#endif

	uint8_t bytesRequired;
	float delta{ 0 };
	uint16_t tempUint16{ 0 };

	delta = maxValue - minValue + 1;
	for (int i = 1; i <= (int)decimals; i++ ) {
		if (divide == false) {
			delta *= (float)10;
		}
		else {
			delta /= (float)10;
		}
	}

	bytesRequired = delta > 254.0 ? 2 : 1;
#ifdef _DEBUG_BFFDC_BYTES_REQUIRED
	PRINT_VALUE_EQUAL_TAB_NL(floatValueName);
	PRINT_VALUE_EQUAL_TAB_NL(bytesRequired);
	PRINT_VALUE_EQUAL_TAB_NL(divide);
	PRINT_VALUE_EQUAL_TAB_NL(decimals);
	PRINT_VALUE_EQUAL_TAB_NL(delta);
#endif

#ifdef _DEBUG_BFFDC_BYTES_REQUIRED
	PRINT_FUNCTION_EXIT_NL;
#endif
	return bytesRequired;
} // CalculateBytesRequiredFloat

uint16_t BFFDC::TransformFloatIntoInt16(bool& _errorCondition) {
	// PROGRAM DOC START TransformFloatIntoInt16()
	// Takes a single float value, compacts it, and transforms it into an int16_t value.
	// This function needs to be called repeatedly for all data values.
	// TransformFloatIntoInt16()and TransformInt16IntoFloat() work in tandem
	// Float comes in as raw sensor value, for example humidity would be 68%
	// check negatives allowed
	// Check min/max
	// Effect transformation as per data control then convert
	// according to parameter data to an uint32_t
	// then put data in local byte array and set negative allowed bit to 1 if required (MSB bit 6)
	// Once this is done, transfer to global array at the correct offset
	// Gets decimals and array size from sensor's control data (dataBytesFromFloatsSensorsTable[])
	// ERROR CHECKING
	// Error checks total : 7
	// If float is negative, checks that it is allowed
	// Checks getting control variables in one throw via GetAllControlParameters()
	// Checks float to uint32 conversion call function
	// Checks that value is within range
	// // PARAMETERS
	// Parameters total : 1
	// uint8_t _dataIndex ??
	// RETURNS
	// Returns bool - success/failure
	// PROGRAM DOC END TransformFloatIntoInt16 ()
	// PROGRAM FLOW START calling function: TransformFloatIntoInt16 ()

#ifdef _DEBUG_BFFDC_TRANSFER_FLOAT_TO_INT16_ARRAY
	;
	PRINT_VALUE_EQUAL_DESC("Start: ", floatValueName);
	PRINT_VALUE_EQUAL_DESC_NL(" = ", floatValue);
#endif

	float uint16Value{ floatValue };
#ifdef _DEBUG_BFFDC_TRANSFER_FLOAT_TO_INT16_ARRAY
	PRINT_VALUE_EQUAL_DESC_NL("Just started ", uint16Value);
#endif

	// Adjust value to zero base
	uint16Value -= minValue; // Adjust value to zero base
#ifdef _DEBUG_BFFDC_TRANSFER_FLOAT_TO_INT16_ARRAY
	PRINT_VALUE_EQUAL_DESC_NL("Ajusted zero base ", uint16Value);
#endif

	// modify value as per multiplier if required
	// multiplier and decimals are in units of 10.0, for 0  multiply 1 (aka do not multiply)
	uint8_t decimals = ExtractValueFromBitmap(EControlVariables::decimals);
#ifdef _DEBUG_BFFDC_TRANSFER_FLOAT_TO_INT16_ARRAY
	PRINT_VALUE_EQUAL_TAB(decimals);
#endif

#ifdef _DEBUG_BFFDC_TRANSFER_FLOAT_TO_INT16_ARRAY
	PRINT_VALUE_EQUAL_DESC_NL("", uint16Value);
#endif

	bool doDivideInstead = ExtractValueFromBitmap(EControlVariables::divide);
	//Multiply or divide
	for (int i = 1; i <= decimals; i++) {
		if (doDivideInstead == false) {
			// do multiplication
			uint16Value *= 10;
		}
		else { // do division
			uint16Value /= 10;
		}
	} // for (int i = 1; i < decimals; i++)

#ifdef _DEBUG_BFFDC_TRANSFER_FLOAT_TO_INT16_ARRAY
	PRINT_VALUE_EQUAL_DESC_NL("Did multiply/divide ", uint16Value);
	PRINT_VALUE_EQUAL_TAB_NL(uint16Value);
#endif

#ifdef _DEBUG_BFFDC_TRANSFER_FLOAT_TO_INT16_ARRAY
	PRINT_FUNCTION_EXIT_NL
#endif
		return uint16Value;
	// PROGRAM FLOW END calling function: TransformFloatIntoInt16()
} // TransformFloatIntoInt16

float BFFDC::TransformInt16IntoFloat() {
	// PROGRAM DOC START TransformInt16IntoFloat ()
	// TransformFloatIntoInt16()and TransformInt16IntoFloat() work in tandem
	// Transfers data in global array based on sensor offset into local array
	// value is negative is in bit6 of MSB (the last byte)
	// Converts to uint32_t then to float. Sets to negative if required
	// Gets decimals and array size from sensor's control data (dataBytesFromFloatsSensorsTable[])
	// ERROR CHECKING
	// Error checks total : 0
	// None
	// PARAMETERS
	// Parameters total : 1
	// uint8_t _dataIndex ??
	// RETURNS
	// Returns float value
	// PROGRAM DOC END TransformInt16IntoFloat ()
	// PROGRAM FLOW START calling function: TransformInt16IntoFloat ()

#ifdef _DEBUG_BFFDC_TRANSFER_INT16_TO_FLOAT
	PRINT_FUNCTION_ENTER_NL;
	PRINT_VALUE_EQUAL_DESC("Start: ", sensorLabels[_dataIndex]);
	PRINT_VALUE_EQUAL_DESC_NL(" = ", _floatValue);
#endif

	uint16_t temp16int{ uint16Value  };
	float tempFloat = (float)temp16int;

	bool doDivideInstead = ExtractValueFromBitmap(EControlVariables::divide);
	//Multiply or divide
	for (int i = 1; i <= decimals; i++) {
		if (doDivideInstead == false) {
			// do multiplication
			tempFloat /= (float)10;
		}
		else { // do division
			tempFloat *= (float)10;
		}
	} // for (int i = 1; i < decimals; i++)

	tempFloat += minValue;


#ifdef _DEBUG_BFFDC_TRANSFER_INT16_TO_FLOAT
	PRINT_FUNCTION_EXIT_NL
#endif
		return tempFloat;
	// PROGRAM FLOW END calling function: TransformInt16IntoFloat()
} // TransferInt32ToFloat()

// Begin BFFDC.cpp
