#ifndef BFFDC_H
#define BFFDC_H

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// incorporate it inside of an executable or other machine language component
// without restriction for personal or commercial use.The Software may not be
// redistributed as source code or in any recognizable human readable form
// in any form whatsoever for any use whatsoever.
// 
// The Software is the property of Michle Delsol(France), copyright 2023 USA and
// international. For any questions concerning use of the software contact
// Michle Delsol at CPParduino@md-dsl.fr.
// 
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include "Globals.h"

// PROGRAM FILE SET START fileName BFFDC.h

// Table Driven Number Data Transfer Framework
// The BFFDC (Byte to Float Format Driven Conversion) class aims to reduce the memory size required 
// to store float data. Note that the class operates on one float value at a time. if an array of 
// floats is to be handled, the class's functions must be called repeatedly, once for each value.
// This class differs from the DataGroups class in that data is obtained
// as float values. In a typical scenario such as a beehive weighing system, six sensor float values 
// require 24 bytes to store the data. The BFFDC class can reduce this requirement to 4 or 6 bytes.
// The BFFDC class provides a set of methods to enable transparent sensor data to be transferred
// as a stream of bytes.
// A float is encoded into a byte array, transmitted, and decoded at the receiving end into a float.
// Two distinct issues are handled by this class:
//   It provides the mechanics for transforming a float data value into a byte array and vice versa.
//   Since range and precision may vary from one data value to another, each data value is customize 
//   to respond to application needs (dataValueControlOperators).
// For example, a temperature can range from 20C to +60C, no decimals.
// This means it will fit in one byte, no decimals, signed bit on, no multiplication.
// Weights will range from 0.0V to 5.0V (conversion to kgs is effected after data transfer)
// which translates as one byte, 1 decimal, sign byte off, multiplier at 1 (multiply by 10).
// This means that data conversion will be controlled by a set of parameters.
// A sensor parameter table accessed by a sensor enum, enables
// the appropriate sensor parameter to be extracted from the table.

// The algorithm works as follows:
// On the send end:
// 1 - The sensor value is adjusted to zero base by subtracting from its value the minimum range (sensorRange)
// 2 - It is multiplied by 10^decimals to transform it into an unsigned integer (uint16_t), or divided if the divide instead bit is on.
// 3 - If it is to become a two-byte target, the uint16_t value is separated into LSB/MSB-least-most-significant bytes).
// On the receive end
// 1 - If it is to become a two-byte value, the LSB/MSB-least-most-significant bytes are extracted and assembled into one byte.
// 2 - The result is assigned to a float adn divided by 10^decimals to transform it into an unsigned integer (uint16_t), or multiplied if the divide instead bit is on.
// 3 - The sensor value is adjusted by adding the minimum range (sensorRange)
// 
// Here are the specifics of the sensor data to byte array conversion and vice versa.
// A sensor data item is characterized by
//   the number of decimals (multiplier),
//   by its sign (positive or negative) and
//   by its absolute value when the number is shifted right by the appropriate number of decimals.
//   a multiplier value
//   a divide instead to multiply value
// Difference between decimals and multiplier:
//     decimals - algorithm does  appropriate conversions to maintain the number of significant digits.
//     multiplier will change the input value to reduce the data size being transmitted. For example, for a
//     humidity of 58%, round of to 60%, remove 30% because humidity never drops below this value,
//     divide by 10 to get 3 and transmit. At the receiving end,
//     multiply by 10 and add 30 to display 60% which is deemed good enough.
// The methods transform the float into a uint32_t value (unsigned 32 bit integer)
// which is then broken down into the appropriate number of bytes for data transfer.
// When transferring data, 255 values are discarded due to the fact that
// stray 255 bytes are introduced into the data stream. Hence only 7 of the 8 bits are used.
// When handling negative numbers, the last usable bit, bit 6, is used to signal that
// the number is negative. Since 0 based, bit 6 is the 7th bit).
// If the float converted to uint32_t is less than 63 positive or negative, one byte suffices.
// For larger numbers, two or more bytes are required hence there are two types of bytes:
//   MSB byte, the last one, contains the signed value, bit 6 to signal negative
//   LSB bytes, contains values defined by bits 0 to 5 inclusive
// For small numbers, the LSB and MSB are one and the same. The LSB is an MSB.
// Once the uint32_t value and sign have been defined, the methods transform
// the uint32_t value into a sequence of 7 bits bytes for data transfer or reverse.

// The first order of business is to define the characteristics of data values. Here is the one I defined for
// the hive sensors:
// item        range    bytes   decimals negatives  multiplier  divide instead
// 8V        0.0 to 9.0     2        1      false         0            0     // No need for decimals
// 5V        0.0 to 5.0     1        1      false         0            0     // No need for decimals
// temp      -40 to  60     1        0      true          0            0     // No need for decimals
// humid       0 to 100     1        0      false         1            1     // 40%, 50%... sufficient
// weights - for 5 volts equal 100 kgs, one decimal, 50 values, allows a 2.0kg resolution, two decimals, 500 values, 0.20kg which is deemed enough
// body     0.00 to 5.00    3        2      false         0            0     // 4.34V, 4.46V... no need for 2nd decimal
// super  0.00 to 5.00      3        2      false         0            0     // 4.34V, 4.46V... no need for 2nd decimal

// The next order of business is to encode data value characteristics contained in a uint8_t table. These are 
// bit size operators. 
// These operators fit inside one byte, their structure is:
// 
// bits 0-1 bytes - number of bytes: 1 to 3
// bits 2-3 decimals - number of decimals; 0 to 3 decimal places
// bit 4 - negative allowed: if 1 true
// bits 5-6 multiplier: final value is multiplied by 10*multiplier when sending, divided 10 times when receiving
// bit 7 - divide or multiply: 0 is multiply, 1 is divide
// 
// Since the application will most probably need to handle several data values (sensors), 
// these operators are located in a table: dataBytesFromFloatsSensorsTable

// Since the sensor data control operators are individual bits inside a register, 
// the info is extracted by appropriate masking and shifting the result, thus:
// extraction:  value &= mask // where mask could be 00110000 ? bits 2 and 3 extracted
// shift: value >> 4 // to get the result.
// Since there are 5 control items, an enum driven table define these:
//   uint8_t dataValueControlOperators[]  {
//      { B11000000 }, // 2 bits - NumberBytes thus 3 bytes max
//      { B00110000 }, // 2 bits - NumberDecimals
//      { B00001000 }, // 1 bit - AllowNegatives
//      { B00000110 }, // 2 bits - Multiplier
//      { B00000001 }  // 1 bit - DivideInstead
//    ;
// For each control item in the above, a mask is defined along with the number of bytes
// used by the control item. Thus, the control item for the number of bytes of any sensor data item
// has a mask B11000000 which occupies 2 bytes. This means that the raw float sensor
// value of any item can be put into a max of 3 bytes. This value is obtained
// by doing an & mask and shifting by the cumulative values of controls value byte sizes.
// Thus, to extract number of decimals:
//     & mask with B00110000
//      shift 8 - 2+2 = 4 to the right.
//
// One next order of business is to define a range of values dor data values (sensor range). Example
// Min/max to reduce mem requirements  - these are referred to as uint8_t minMaxRanges
//uint8_t minMaxValues[][2] = { // Consumes 48 bytes
//  { 0, 9}, // Voltage8V
//  { 0, 6}, // Voltage5V
//  { 4, 60}, // Temp
//  { 0, 10}, // Humid - in tens thus 10 is really 100
//  { 0, 5}, // WeightBody
//  { 0, 5}, // WeightSuper
//};


// Table of max int32_t values
//            bytes  positive only  neg allowed
// Max Value      1       127            63
// Max Value      2     16383          8191
// Max Value      3   2097151       1048575
// Max Value      4 268435455     134217727

// PROGRAM FILE SET END fileName BFFDC.h

// Global
enum class EControlVariables : uint8_t {start = 0, decimals = 0, divide, end};

//*************************************************************************************

// Support class to handle array compresoin/expansion
class UAC { // short for UintArrayCompaction
public:
	UAC(uint8_t _noItems,
		uint8_t _bytesRequiredTotal, uint16_t* _rawUint16data,
		uint8_t* _bytesRequiredIndividualArray, uint8_t* _compactedUint8Array);
	void ProcessCompactToRaw();
	void ProcessRawToCompact();
private:
	uint8_t noItems; // defies array size
	uint8_t bytesRequiredTotal;
	// These array pointers are brought in from the outside
	// Array values once the conversion process has finished
	uint16_t* rawUint16data;
	// Most variables, once compacted, require one byte only; some require two
	uint8_t* bytesRequiredIndividualArray;
	// Array to be used in data transfer, oversized - bytesRequiredTotal defines usable length
	uint8_t* compactedUint8Array;
}; // class UAC

//*************************************************************************************
// Convert float value to compacted uint value
class BFFDC {
public:
	// Constructor gets float value, generates uint16_t value
	BFFDC(bool& _errorCondition, float _floatValue, float _minMaxValues[], uint8_t _controlVariables, char* _floatValueName);
	// Constructor gets uint16_t value, generates float value
	BFFDC(bool& _errorCondition, uint16_t _uint1Value, float _minMaxValues[], uint8_t _controlVariables, char* _floatValueName);
	BFFDC() {};
	~BFFDC();
	uint16_t TransformFloatIntoInt16(bool& _errorCondition);
	float TransformInt16IntoFloat();
	uint8_t BytesRequired();
	uint8_t GetVarMask(EControlVariables _varID) { return controlVariableMasks[(uint8_t)_varID]; };
	uint8_t GetMaskOffset(EControlVariables _varID) { return controlVariableMaskOffset[(uint8_t)_varID]; };
	char* GetVarLabel(EControlVariables _varID) { return controlLabels[(uint8_t)_varID]; };
	uint8_t ExtractValueFromBitmap(EControlVariables _varID);
	void PrintValuesInControlVariables(); // For debugging
	float minValue{ 0 }; //These two are set in the constructor
	float maxValue{ 0 };
private:
	char* floatValueName{""};
	uint8_t decimals{ 0 };
	bool divide{ false };
	float floatValue; // Value to be transformed to unit8_t array ( 1 or 2 bytes)
	uint16_t uint16Value; // Value to be transformed to unit8_t array ( 1 or 2 bytes)
	uint8_t controlVariables;
	static char* controlLabels[];
	static uint8_t controlVariableMasks[];
	static uint8_t controlVariableMaskOffset[];
}; // class

#endif // BFFDC_H
