// Begin HelperFunctions.cpp

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// incorporate it inside of an executable or other machine language component
// without restriction for personal or commercial use.The Software may not be
// redistributed as source code or in any recognizable human readable form
// in any form whatsoever for any use whatsoever.
// 
// The Software is the property of Michèle Delsol(France), copyright 2023 USA and
// international. For any questions concerning use of the software contact
// Michèle Delsol at CPParduino@md-dsl.fr.
// 
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include "Globals.h"

//------------------------------STRING NUMERIC CONVERSION ---------------------

bool MaxPossibleValueInNnumberOfBytes(uint32_t* _maxPossible32int, uint8_t _allocatedByteNumber, uint8_t _bitSize, char* _callingFunction, int _lineNo) {
    // FUNCTION DESCRIPTION MaxPossibleValueInNnumberOfBytes()
    // Gets the maximum possible value based on number of bytes allocated:
    // Should return one of the following in _maxPossible32int:
    // Since
    // Table of max int32_t values
    //            bytes  positive only  neg allowed
    // Max Value      1       127            63
    // Max Value      2     16383          8191
    // Max Value      3   2097151       1048575
    // Max Value      4 268435455     134217727
    // Since a negative allowed bit is always set on the MSB byte,
    // the max possible value shall always be the lesser of the two above,
    // i.e. the one corresponding to a negative value since we are losing 1 bit
    // to define the number being transmitted
    // ERROR CHECKING
    // Error checks total : 2
    // checks the number of: bytes 1 to 4
    // Check that bitSize is 7, 8 for future work, anything else is an error
    // PARAMETERS
    // Parameters total : 3
    // uint32_t* _maxPossible32int - max value possible for byte array size equal to _byteNumber
    // uint8_t _allocatedByteNumber - number of bytes allocated.
    // uint8_t _bitSize - number of bits to store data. 7 bits is only default, 8 bits for future work, anything elseis an error
    // RETURNS
    // Returns bool - success/failure
    // PROGRAM DOC END MaxPossibleValueInNnumberOfBytes()
    // PROGRAM FLOW START calling function: MaxPossibleValueInNnumberOfBytes()

#ifdef _DEBUG_HELPER_FUNCTION_MAX_POSSIBLE_VALUE_IN_N_NUMBER_OF_BYTES
    PRINT_FUNCTION_ENTER_NL
        PRINT_VALUE_EQUAL_TAB(_byteNumber)
#endif

        // Error check number of bytes
        if (_allocatedByteNumber < 1 || _allocatedByteNumber > 4) {
#if defined(_DEBUG_HELPER_FUNCTION_MAX_POSSIBLE_VALUE_IN_N_NUMBER_OF_BYTES)
            PROGRAM_ERROR_TEXT_VALUE("Number of bytes out of range", _allocatedByteNumber)
                PRINT_FUNCTION_EXIT_NL
#endif
                * _maxPossible32int = 0;
            return false;
        }
    *_maxPossible32int = 1;
    for (int i = 1; i <= _allocatedByteNumber; i++) {
        *_maxPossible32int *= 128L;
#ifdef _DEBUG_HELPER_FUNCTION_MAX_POSSIBLE_VALUE_IN_N_NUMBER_OF_BYTES
        Serial.print(F("i "));
        Serial.print(i);
        Serial.print(F(", *_maxPossible32int "));
        Serial.println(*_maxPossible32int);
#endif
    }

    // Check for bit size
    if (_bitSize == 7) {
        ; // continue as this is allowed
    }
    else if (_bitSize == 8) {
        // Error currently not supported
#if defined (ERROR_REPORTING_ON) || defined(_DEBUG_HELPER_FUNCTIONS_TRANSFER_UINT32_TO_BYTE_ARRAY)
        PROGRAM_ERROR_TEXT("bitsUsed 8 currently not supported")
#endif
            return false;
    }
    else {
        // Error - should be 7 or 8
#if defined (ERROR_REPORTING_ON) || defined(_DEBUG_HELPER_FUNCTIONS_TRANSFER_UINT32_TO_BYTE_ARRAY)
        PROGRAM_ERROR_TEXT("bitsUsed should be 7 or 8")
#endif
            return false;
    }

    float basPow = 128.0;
    if (_bitSize == 8) {
        basPow = 256.0;
    }

    // Divide by 2 since lose one bit for the negative allowed bit.
    *_maxPossible32int >>= 1;

    (*_maxPossible32int)--;
#ifdef _DEBUG_HELPER_FUNCTION_MAX_POSSIBLE_VALUE_IN_N_NUMBER_OF_BYTES
    Serial.print(F("_allocatedByteNumber "));
    Serial.print(_allocatedByteNumber);
    Serial.print(F(", *_maxPossible32int "));
    Serial.println(*_maxPossible32int);
#endif

#ifdef _DEBUG_HELPER_FUNCTION_MAX_POSSIBLE_VALUE_IN_N_NUMBER_OF_BYTES
    PRINT_FUNCTION_EXIT_NL
#endif
        return true;
    // PROGRAM FLOW END calling function: MaxPossibleValueInNnumberOfBytes()
} // MaxPossibleValueInNnumberOfBytes()

bool ConvertFloatToUint32(float _floatValue, uint32_t* _uint32value, char* _callingFunction, int _lineNo) {
    // FUNCTION DESCRIPTION ConvertFloatToUint32 ()
    // Takes a float, the before decimals part, could be negative, and converts it to a positive uint32_t
    // the decimal part of the float is discarded thus for 45.63 only 45 is converted
    // Negative sign is discarded, calling function must handle negative values
    // ERROR CHECKING
    // None
    // Error checks total : 0
    // PARAMETERS
    // Parameters total : 2
    // float _floatValue - Float value to convert, arbitrary number of decimals
    // uint32_t* _uint32value - destination value
    // RETURNS
    // Returns bool � success/failure
    // PROGRAM DOC END ConvertFloatToUint32 ()
    // PROGRAM FLOW START calling function: ConvertFloatToUint32 ()

#ifdef _DEBUG_HELPER_FUNCTION_CONVERT_FLOAT_TO_UINT32
    PRINT_FUNCTION_ENTER_NL
        PRINT_VALUE_EQUAL_TAB(_floatValue)
#endif

        // code starts here
        float tempFloat{ _floatValue };

    //Convert to positive value if negative
    if (tempFloat < 0.0) {
        tempFloat *= (-1.0);
    }

    *_uint32value = (uint32_t)tempFloat;
#ifdef _DEBUG_HELPER_FUNCTION_CONVERT_FLOAT_TO_UINT32
    PRINT_VALUE_EQUAL_TAB(tempFloat)
        PRINT_VALUE_EQUAL_TAB(*_uint32value)
#endif

#ifdef _DEBUG_HELPER_FUNCTION_CONVERT_FLOAT_TO_UINT32
        PRINT_FUNCTION_EXIT_NL
#endif
        return true;
    // PROGRAM FLOW END calling function: ConvertFloatToUint32()
} // ConvertFloatToUint32

bool TransferByteArrayToUint32(uint32_t* _uint32value, uint8_t byteArray[], uint8_t bitsUsed, uint8_t arraySize, bool* isNegative, char* _callingFunction, int _lineNo) {
    // FUNCTION DESCRIPTION TransferByteArrayToUint32()
    // TransferByteArrayToUint32 and TransferUint32ToByteArray work in tandem.
    // TransferUint32ToByteArray receives a uint32_tvalue which will be broken up into 7 or 8 bit chunks
    // TransferByteArrayToUint32 receives a byte array which shall be used to reconstruct value.
    // TransferUint32ToByteArray checks that the array size is compatible with the magnitude of the number to be transfered into the array
    // Sets/gets bit 6 with isNegative
    // ERROR CHECKING
    // Error checks total : 1
    // Checks for minimum heap space to run
    // PARAMETERS
    // Parameters total : 0
    // None
    // RETURNS
    // Returns void
    // PROGRAM DOC END TransferByteArrayToUint32()
    // PROGRAM FLOW START calling function: TransferByteArrayToUint32()

#ifdef _DEBUG_HELPER_FUNCTIONS_TRANSFER_BYTE_ARRAY_TO__UINT32
    PRINT_FUNCTION_ENTER_NL
#endif
#ifdef _DEBUG_HELPER_FUNCTIONS_TRANSFER_BYTE_ARRAY_TO__UINT32
        PRINT_FUNCTION_ENTER_NL
#endif

        // Local variables
        uint32_t maxPossible32int{ 0 };

    // These are not required as value to be gotten is inside the byte array
    //bool functionCallOK = MaxPossibleValueInNnumberOfBytes(&maxPossible32int, arraySize, __FUNCTION__, __LINE__);
    //if (functionCallOK == false) {
    //#if defined (ERROR_REPORTING_ON) || defined(_DEBUG_HELPER_FUNCTIONS_TRANSFER_UINT32_TO_BYTE_ARRAY)
    //PROGRAM_ERROR_TEXT("MaxPossibleValueInNnumberOfBytes problem: false")
    //PRINT_VALUE_EQUAL_TAB(arraySize)
    //#endif
    //}
    //if (maxPossible32int < _uint32value) {
    //// Problem as number does not fit in supplied array
    //#if defined (ERROR_REPORTING_ON) || defined(_DEBUG_HELPER_FUNCTIONS_TRANSFER_UINT32_TO_BYTE_ARRAY)
    //PROGRAM_ERROR_TEXT("Array not large enough for number, try with larger array")
    //PRINT_VALUE_EQUAL_TAB(arraySize)
    //#endif
    //return false;
    //}

    // Check for bit size
    if (bitsUsed == 7) {
        ; // continue as this is allowed
    }
    else if (bitsUsed == 8) {
        // Error currently not supported
#if defined (ERROR_REPORTING_ON) || defined(_DEBUG_HELPER_FUNCTIONS_TRANSFER_UINT32_TO_BYTE_ARRAY)
        PROGRAM_ERROR_TEXT("bitsUsed 8 currently not supported")
#endif
            return false;
    }
    else {
        // Error - should be 7 or 8
#if defined (ERROR_REPORTING_ON) || defined(_DEBUG_HELPER_FUNCTIONS_TRANSFER_UINT32_TO_BYTE_ARRAY)
        PROGRAM_ERROR_TEXT("bitsUsed should be 7 or 8")
#endif
            return false;
    }

    float basPow = 128.0;
    if (bitsUsed == 8) {
        basPow = 256.0;
    }

#ifdef _DEBUG_HELPER_FUNCTIONS_TRANSFER_BYTE_ARRAY_TO__UINT32
    Serial.println(F("Now get value from array"));
#endif
    uint32_t finalValue = 0;
    uint32_t multiplier = 1;

    for (uint32_t i = 0; i < arraySize; i++) {
        if (i > 0)  multiplier *= basPow;
        finalValue += (uint32_t)byteArray[i] * (multiplier);
#ifdef _DEBUG_HELPER_FUNCTIONS_TRANSFER_BYTE_ARRAY_TO__UINT32
        Serial.print(F("(uint32_t)(basPow+0.1) ")); Serial.print((uint32_t)(basPow + 0.1));
        Serial.print(F(", multiplier ")); Serial.print(multiplier);
        Serial.print(F(", arrayStoreValue ")); Serial.print(byteArray[i]);
        Serial.print(F(", finalValue ")); Serial.println(finalValue);
#endif
    }
    *_uint32value = finalValue;

#ifdef _DEBUG_HELPER_FUNCTIONS_TRANSFER_BYTE_ARRAY_TO__UINT32
    PRINT_FUNCTION_EXIT_NL
#endif
        return true;
    // PROGRAM FLOW END calling function: TransferByteArrayToUint32()
} // TransferByteArrayToUint32

bool TransferUint32ToByteArray(uint32_t _uint32value, uint8_t byteArray[], uint8_t bitsUsed, uint8_t arraySize, bool isNegative, char* _callingFunction, int _lineNo) {
    // FUNCTION DESCRIPTION TransferUint32ToByteArray()
    // TransferByteArrayToUint32 and TransferUint32ToByteArray work in tandem.
    // TransferUint32ToByteArray receives a uint32_tvalue which will be broken up into 7 or 8 bit chunks
    // TransferByteArrayToUint32 receives a byte array which shall be used to reconstruct value.
    // TransferUint32ToByteArray checks that the array size is compatible with the magnitude of the number to be transfered into the array
    // Sets/gets bit 6 with isNegative
    // ERROR CHECKING
    // Error checks total : 1
    // Checks for minimum heap space to run
    // PARAMETERS
    // Parameters total : 0
    // None
    // RETURNS
    // Returns void
    // PROGRAM DOC END TransferUint32ToByteArray()
    // PROGRAM FLOW START calling function: TransferUint32ToByteArray()

#ifdef _DEBUG_HELPER_FUNCTIONS_TRANSFER_UINT32_TO_BYTE_ARRAY
    PRINT_FUNCTION_ENTER_NL
#endif

        // Local variables
        uint8_t calculatedArraySize{ 0 };
    uint32_t maxPossible32int{ 0 };

    bool functionCallOK = MaxPossibleValueInNnumberOfBytes(&maxPossible32int, arraySize, 7, __FUNCTION__, __LINE__);
    if (functionCallOK == false) {
#if defined (ERROR_REPORTING_ON) || defined(_DEBUG_HELPER_FUNCTIONS_TRANSFER_UINT32_TO_BYTE_ARRAY)
        PROGRAM_ERROR_TEXT("MaxPossibleValueInNnumberOfBytes problem: false")
            PRINT_VALUE_EQUAL_TAB(arraySize)
#endif
    }
    if (maxPossible32int < _uint32value) {
        // Problem as number does not fit in supplied array
#if defined (ERROR_REPORTING_ON) || defined(_DEBUG_HELPER_FUNCTIONS_TRANSFER_UINT32_TO_BYTE_ARRAY)
        PROGRAM_ERROR_TEXT("Array not large enough for number, try with larger array")
            PRINT_VALUE_EQUAL_TAB(arraySize)
            PRINT_VALUE_EQUAL_TAB(_uint32value)
#endif
            return false;
    }

    // Check for bit size
    if (bitsUsed == 7) {
        ; // continue as this is allowed
    }
    else if (bitsUsed == 8) {
        // Error currently not supported
#if defined (ERROR_REPORTING_ON) || defined(_DEBUG_HELPER_FUNCTIONS_TRANSFER_UINT32_TO_BYTE_ARRAY)
        PROGRAM_ERROR_TEXT("bitsUsed 8 currently not supported")
#endif
            return false;
    }
    else {
        // Error - should be 7 or 8
#if defined (ERROR_REPORTING_ON) || defined(_DEBUG_HELPER_FUNCTIONS_TRANSFER_UINT32_TO_BYTE_ARRAY)
        PROGRAM_ERROR_TEXT("bitsUsed should be 7 or 8")
#endif
            return false;
    }

    float basPow = 128.0;
    if (bitsUsed == 8) {
        basPow = 256.0;
    }

    uint32_t multiplier = 1;
#ifdef _DEBUG_HELPER_FUNCTIONS_TRANSFER_UINT32_TO_BYTE_ARRAY
    Serial.print(F("Using base ")); Serial.println(basPow);
#endif
    for (uint8_t i = 0; i < arraySize; i++) {
        float powResult = pow(basPow, i) + 0.1;
        uint32_t tempValue = (uint32_t)powResult;
        tempValue *= basPow;
#ifdef _DEBUG_HELPER_FUNCTIONS_TRANSFER_UINT32_TO_BYTE_ARRAY
        Serial.print(F("i ")); Serial.print(i);
        Serial.print(F(", basPow ")); Serial.print(powResult);
        Serial.print(F(", tempValue ")); Serial.print(tempValue);
#endif
        byteArray[i] = (_uint32value % tempValue) / (tempValue / basPow);
#ifdef _DEBUG_HELPER_FUNCTIONS_TRANSFER_UINT32_TO_BYTE_ARRAY
        Serial.print(F(", byteArray ")); Serial.println(byteArray[i]);
#endif
    }

#ifdef _DEBUG_HELPER_FUNCTIONS_TRANSFER_UINT32_TO_BYTE_ARRAY
    PRINT_FUNCTION_EXIT_NL
#endif
        return true;
    // PROGRAM FLOW END calling function: Aardvark()
} // TransferUint32ToByteArray

// End HelperFunctions.cpp
