// Begin BitfieldStorage.cpp

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// incorporate it inside of an executable or other machine language component
// without restriction for personal or commercial use.The Software may not be
// redistributed as source code or in any recognizable human readable form
// in any form whatsoever for any use whatsoever.
// 
// The Software is the property of Michle Delsol(France), copyright 2023 USA and
// international. For any questions concerning use of the software contact
// Michle Delsol at CPParduino@md-dsl.fr.
// 
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.


// Scenario:
// The BitfieldStorage Framework provides a class which contains many data items
// which require less than 8 bits, such as flags which can be true or false, thus
// requiring one bit only, to be stored as vit fields and accesses, modified via
// generic functions which use an enum list to identified the variable to be worked on.

#include "Globals.h"

//------------------------- BITFIELD STORAGE -------------------------------
#define _DEBUG_BITFIELD_STORAGE_ON
#ifdef _DEBUG_BITFIELD_STORAGE_ON

#define _DEBUG_BITFIELD_STORAGE_BITFIELD_STORAGE
//#define _DEBUG_BITFIELD_STORAGE_VALUES_ARRAY_COHERENCE
//#define _DEBUG_BITFIELD_STORAGE_CREATE_MASK
#define _DEBUG_BITFIELD_STORAGE_SET_BITFIELD_VALUE
//#define _DEBUG_BITFIELD_STORAGE_GET_BITFIELD_VALUE
//#define _DEBUG_BITFIELD_STORAGE_GET_OFFSET_IN_MAP
#define _DEBUG_BITFIELD_STORAGE_VERIFY_MASKS
#define _DEBUG_BITFIELD_STORAGE_
#define _DEBUG_BITFIELD_STORAGE_
#define _DEBUG_BITFIELD_STORAGE_PRINT_VARIABLES
#define _DEBUG_BITFIELD_STORAGE_PRINT_MASKS


#endif // _DEBUG_BITFIELD_STORAGE_ON

BitfieldStorage::BitfieldStorage(uint8_t* _valuesArray, uint8_t _varCount)
	: valuesArray(_valuesArray), varCount(_varCount) {

	// Constructor. Checks storage array coherence via ValuesArrayCoherence()
	// Parameters
	// uint8_t* _valuesArray - array of uint8_t variables passed to the constructor. 
	// uint8_t _varCount - the total number of variables
	// VarCharacteristics* _varCharacteristics - characteristics individual variables

	uint8_t faultyIndex{ 0 };

#ifdef _DEBUG_BITFIELD_STORAGE_BITFIELD_STORAGE
	PRINT_FUNCTION_ENTER_NL
		PRINT_VALUE_EQUAL_NL(varCount);
#endif

	boolean valuesArrayOK = ValuesArrayCoherence(faultyIndex);
	if (valuesArrayOK == false) {
#ifdef _DEBUG_BITFIELD_STORAGE_BITFIELD_STORAGE
		Serial.print(F("Problem with bit size at item "));
		Serial.print(faultyIndex);
		Serial.print(F(", "));
		Serial.println(varCharacteristics[faultyIndex].label);
#endif
		return;
	} // if

#ifdef _DEBUG_BITFIELD_STORAGE_BITFIELD_STORAGE
	for (int i = 0; i < _varCount; i++) {
		uint8_t byteNo{ 0 };
		uint8_t offsetInMap = GetOffsetInMap(i);
		uint8_t offsetInSequence = GetOffsetInSequence(i);
		if (offsetInMap % 8 == 0) {
			PRINT_VALUE_EQUAL_NL(byteNo);
			byteNo++;
		}
		Serial.print(F("\ti = "));
		Serial.print(i);
		Serial.print(F(", "));
		Serial.print(varCharacteristics[i].label);
		Serial.print(F(", bitsize = \t"));
		Serial.print(varCharacteristics[i].size);
		Serial.print(F(", offsetInMap = \t"));
		Serial.print(offsetInMap);
		Serial.print(F(", offsetInSequence = \t"));
		Serial.println(offsetInSequence);
	}
#endif

#ifdef _DEBUG_BITFIELD_STORAGE_BITFIELD_STORAGE
	PRINT_FUNCTION_EXIT_NL
#endif
}; // BitfieldStorage constructor

//-------------------------------- GET/SET FUNCTIONS -------------------------------
void BitfieldStorage::Set(uint8_t _itemNumber, uint8_t _valueToSet) {
#ifdef _DEBUG_BITFIELD_STORAGE_SET_BITFIELD_VALUE
	Serial.print(F("Entering BitfieldStorage::")); Serial.println(__FUNCTION__);
	Serial.print(F("Setting value for _itemNumber =  "));
	Serial.print(_itemNumber);
	Serial.print(F("  "));
	Serial.print(varCharacteristics[_itemNumber].label);
	Serial.print(F(", _valueToSet = "));
	Serial.println(_valueToSet);
#endif
	int8_t regIndex = GetSequenceIndex(_itemNumber);
	uint8_t offsetInMap = GetOffsetInMap(_itemNumber);
	uint8_t offsetInSequence = GetOffsetInSequence(_itemNumber);
	uint8_t bitSize = GetItemBitSize(_itemNumber);
	uint8_t offsetEnd{ offsetInSequence + bitSize - 1 };
	uint8_t indexInRegister{ offsetInMap - (8 * regIndex) };

#ifdef _DEBUG_BITFIELD_STORAGE_SET_BITFIELD_VALUE
	Serial.print(F("Entering BitfieldStorage::")); Serial.println(__FUNCTION__);
	Serial.print(F("regIndex = ")); Serial.println(regIndex);
	Serial.print(F("offsetInSequence = ")); Serial.println(offsetInSequence);
	Serial.print(F("bitSize = ")); Serial.println(bitSize);
	Serial.print(F("offsetEnd = ")); Serial.println(offsetEnd);
	Serial.print(F("indexInRegister = ")); Serial.println(indexInRegister);
#endif

	// From here on all parameters should be OK
	uint8_t mask{ 0 };
	CreateMask(_itemNumber, (uint8_t&)mask);
#ifdef _DEBUG_BITFIELD_STORAGE_SET_BITFIELD_VALUE
	Serial.print(F("\tmask ")); Serial.println(mask, BIN);
#endif

	// Now do work
	// Simulate one byte at offset into the mapped values
	uint8_t* mappedBytes = valuesArray;
	// Down to here Get and Set are identical

	uint8_t tempValue{ _valueToSet };
	tempValue <<= (offsetInSequence - bitSize + 1);
	mappedBytes[regIndex] &= ~mask;
	mappedBytes[regIndex] += tempValue;

#ifdef _DEBUG_BITFIELD_STORAGE_SET_BITFIELD_VALUE
	Serial.println(F("After processing "));
	Serial.print(F("Address valuesArray ")); Serial.println((long)valuesArray);
	Serial.print(F("Address mappedBytes ")); Serial.println((long)mappedBytes);
	Serial.print(F("regIndex ")); Serial.print(regIndex);
	Serial.print(F(" value content ")); Serial.print(mappedBytes[regIndex]);
	Serial.print(F(" bin content ")); Serial.println(mappedBytes[regIndex], BIN);
#endif
	return;
} // SetBitfieldStorageItemValue()

uint8_t BitfieldStorage::Get(uint8_t _itemNumber) {
#ifdef _DEBUG_BITFIELD_STORAGE_GET_BITFIELD_VALUE
	Serial.print(F("Entering BitfieldStorage::")); Serial.println(__FUNCTION__);
	Serial.print(F("Getting value for _itemNumber =  "));
	Serial.print(_itemNumber);
	Serial.print(F("  "));
	Serial.println(varCharacteristics[_itemNumber].label);
#endif
	uint8_t valueToGet;
	int8_t regIndex = GetSequenceIndex(_itemNumber);
	uint8_t offsetInMap = GetOffsetInMap(_itemNumber);
	uint8_t offsetInSequence = GetOffsetInSequence(_itemNumber);
	uint8_t bitSize = GetItemBitSize(_itemNumber);
	uint8_t offsetEnd{ offsetInSequence + bitSize - 1 };
	uint8_t indexInRegister{ offsetInMap - (8 * regIndex) };

#ifdef _DEBUG_BITFIELD_STORAGE_GET_BITFIELD_VALUE
	Serial.print(F("regIndex = ")); Serial.println(regIndex);
	Serial.print(F("offsetInSequence = ")); Serial.println(offsetInSequence);
	Serial.print(F("bitSize = ")); Serial.println(bitSize);
	Serial.print(F("offsetEnd = ")); Serial.println(offsetEnd);
	Serial.print(F("indexInRegister = ")); Serial.println(indexInRegister);
#endif

	// From here on all parameters should be OK
	uint8_t mask{ 0 };
	CreateMask(_itemNumber, mask);
#ifdef _DEBUG_BITFIELD_STORAGE_GET_BITFIELD_VALUE
	Serial.print(F("\tmask ")); Serial.println(mask, BIN);
#endif

	// Now do work
	// Simulate one byte at offset into the mapped values
	uint8_t* mappedBytes = valuesArray;

	uint8_t tempValue{ 0 };
	tempValue = mappedBytes[regIndex];
	//Serial.print(F("Just before applying mask tempValue = ")); Serial.println(tempValue);
	tempValue &= mask;
	//Serial.print(F("Just after applying mask tempValue = ")); Serial.println(tempValue);
	tempValue >>= (offsetInSequence - bitSize + 1);
	valueToGet = tempValue;

#ifdef _DEBUG_BITFIELD_STORAGE_GET_BITFIELD_VALUE
	Serial.print(F("Exiting ")); Serial.println(__FUNCTION__);
#endif
	return valueToGet;
} // GetBitfieldStorageItemValue()

// ------------------------------ AUXILIARY PUBLIC FUNCTIONS -------------------------
boolean BitfieldStorage::VerifyMasks() {
#ifdef _DEBUG_BITFIELD_STORAGE_VERIFY_MASKS
	Serial.print(F("Entering Storage::")); Serial.println(__FUNCTION__);
#endif

	uint8_t totalSequenceMask{ 0 };
	uint8_t sequenceCount = (GetOffsetInMap(varCount - 1) + 1) / 8;
	uint8_t sequence{ 0 };
	for (int i = 0; i <= varCount; i++) {
		uint8_t offsetInSequence = GetOffsetInSequence(i);
		if (offsetInSequence == 0) {
			if (i > 0) {
				if (totalSequenceMask != 255) {
#ifdef _DEBUG_BITFIELD_STORAGE_GET_BITFIELD_VALUE
					Serial.print(F("ERROR before item ")); Serial.print(i);
					Serial.print(F(" ")); Serial.print(varCharacteristics[i].label);
#endif
				} // if
#ifdef _DEBUG_BITFIELD_STORAGE_GET_BITFIELD_VALUE
				Serial.print(F("TotalMaskSequence = ")); Serial.println(totalSequenceMask, BIN);
#endif
			}
			if (i < varCount) {
#ifdef _DEBUG_BITFIELD_STORAGE_GET_BITFIELD_VALUE
				Serial.print(F("sequence = ")); Serial.println(sequence);
#endif
			}
			if (i == varCount) break;
			totalSequenceMask = 0;
			sequence++;
		}
		uint8_t mask{ 0 };
		mask = CreateMask(i, mask);
		uint8_t bitSize = GetItemBitSize(i);
		totalSequenceMask += mask;
#ifdef _DEBUG_BITFIELD_STORAGE_PRINT_MASKS
		Serial.print(varCharacteristics[i].label);
		Serial.print(F(" mask = "));
		Serial.print(mask, BIN);
		Serial.print(F(" maskSequence  = ")); Serial.println(totalSequenceMask, BIN);
#endif
	}// for

#ifdef _DEBUG_BITFIELD_STORAGE_PRINT_MASKS
	Serial.print(F("Exiting BitfieldStorage::")); Serial.println(__FUNCTION__);
#endif
	return true;
} // VerifyMasks()

void BitfieldStorage::PrintMasks() {
#ifdef _DEBUG_BITFIELD_STORAGE_PRINT_MASKS
	PRINT_FUNCTION_ENTER_NL
#endif
		for (int i = 0; i < varCount; i++) {
			uint8_t mask{ 0 };
			mask = CreateMask(i, mask);
			uint8_t bitSize = GetItemBitSize(i);
			uint8_t offsetInSequence = GetOffsetInSequence(i);
			uint8_t sequence{ 0 };
			if (offsetInSequence == 0) {
#ifdef _DEBUG_BITFIELD_STORAGE_PRINT_MASKS
				Serial.print(F("sequence = "));
				Serial.println(sequence++);
#endif
			}
#ifdef _DEBUG_BITFIELD_STORAGE_PRINT_MASKS
			Serial.print(F("i = "));
			Serial.print(i);
			Serial.print(F("\t"));
			Serial.print(varCharacteristics[i].label);
			Serial.print(F(" offsetInSequence = "));
			Serial.print(offsetInSequence);
			Serial.print(F(" bitSize = "));
			Serial.print(bitSize);
			Serial.print(F(" mask = "));
			Serial.println(mask, BIN);
#endif
		} // for

#ifdef _DEBUG_BITFIELD_STORAGE_PRINT_MASKS
	PRINT_FUNCTION_EXIT_NL
#endif
		return;
} // PrintMasks()

void BitfieldStorage::PrintVariables() {
	// 
#ifdef _DEBUG_BITFIELD_STORAGE_PRINT_VARIABLES
	PRINT_VALUE_EQUAL_DESC_NL("Entering BitfieldStorage::PrintVariables", varCount);
#endif
	for (int i = 0; i < varCount; i++) {
		uint8_t itemValue{ 0 };
		uint8_t byteNo = GetSequenceIndex(i);
		uint8_t offsetInMap = GetOffsetInMap(i);
		itemValue = Get(i);
		if (offsetInMap % 8 == 0) {
#ifdef _DEBUG_BITFIELD_STORAGE_PRINT_VARIABLES
			Serial.print(F("byteNo = ")); Serial.println(byteNo);
			byteNo++;
#endif
		}
#ifdef _DEBUG_BITFIELD_STORAGE_PRINT_VARIABLES
		Serial.print(F("\ti = "));
		Serial.print(i);
		Serial.print(F("\t"));
		Serial.print(varCharacteristics[i].label);
		PRINT_VALUE_EQUAL_TAB_NL(itemValue);
#endif
	} // for
#ifdef _DEBUG_BITFIELD_STORAGE_PRINT_VARIABLES
	PRINT_FUNCTION_EXIT_NL
#endif
		return;
} // PrintVariables()

//-------------------------------- AUXILIARY PRIVATE FUNCTIONS ---------------------------------
boolean BitfieldStorage::ValuesArrayCoherence(uint8_t& _faultyIndex) {
	// Checks that bitfields fill 8-bit sequences.
	// Returns false if there is a problem
#ifdef _DEBUG_BITFIELD_STORAGE_VALUES_ARRAY_COHERENCE
	PRINT_FUNCTION_ENTER_NL
#endif
		_faultyIndex = 0;
	uint8_t bits{ 0 };
	for (int i = 0; i < varCount; i++) {
#ifdef _DEBUG_BITFIELD_STORAGE_VALUES_ARRAY_COHERENCE
		Serial.print(F("i = "));
		Serial.print(i);
		Serial.print(F(", "));
		Serial.print(varLabels[i]);
#endif
		bits += varCharacteristics[i].size;
		if (bits >= 8) {
			if (bits == 8) {
				bits = 0;
#ifdef _DEBUG_BITFIELD_STORAGE_VALUES_ARRAY_COHERENCE
				Serial.println(F(" : OK"));
#endif
			}
			else {
				// error
#ifdef _DEBUG_BITFIELD_STORAGE_VALUES_ARRAY_COHERENCE
				Serial.println(F(" : BAD"));
#endif
				_faultyIndex = i;
				return false;
			} // if
		}
		else {// if
#ifdef _DEBUG_BITFIELD_STORAGE_VALUES_ARRAY_COHERENCE
			Serial.println(F(" : continue"));
#endif
		}
	} // for
#ifdef _DEBUG_BITFIELD_STORAGE_VALUES_ARRAY_COHERENCE
	PRINT_FUNCTION_EXIT_NL
#endif
		return true;
} // ValuesArrayCoherence()

uint8_t BitfieldStorage::GetOffsetInMap(uint8_t _itemNumber) {
#ifdef _DEBUG_BITFIELD_STORAGE_GET_OFFSET_IN_MAP
	PRINT_FUNCTION_ENTER_NL
#endif
		if (_itemNumber == 0) return 0;
	uint8_t offset{ 0 };
	for (int i = 1; i <= _itemNumber; i++) {
		offset += varCharacteristics[i - 1].size;
		continue;
#ifdef _DEBUG_BITFIELD_STORAGE_GET_OFFSET_IN_MAP
		Serial.print(F("i = "));
		Serial.print(i);
		Serial.print(F(", bitSize i-1 = "));
		Serial.print(varCharacteristics[i - 1].size);
		Serial.print(F(", offset = \t"));
		Serial.println(offset);
#endif
	}
#ifdef _DEBUG_BITFIELD_STORAGE_GET_OFFSET_IN_MAP
	PRINT_FUNCTION_EXIT_NL
#endif
		return offset;
}

uint8_t BitfieldStorage::GetSequenceIndex(uint8_t _itemNumber) {
	return GetOffsetInMap(_itemNumber) / 8;
}

uint8_t BitfieldStorage::GetOffsetInSequence(uint8_t _itemNumber) {
	return GetOffsetInMap(_itemNumber) % 8;
}

uint8_t BitfieldStorage::GetItemBitSize(uint8_t _itemNumber) {
	return varCharacteristics[_itemNumber].size;
} // GetBitSize()

uint8_t BitfieldStorage::CreateMask(uint8_t _itemNumber, uint8_t& _mask) {
	// Creates a mask of type 00011000, position = 3, maskLength = 2
	// When printing get 11000
	// returns mask
#ifdef _DEBUG_BITFIELD_STORAGE_CREATE_MASK
	PRINT_FUNCTION_ENTER_NL
		Serial.print(F("Creating mask for _itemNumber =  "));
	Serial.print(_itemNumber);
	Serial.print(F("  "));
	Serial.println(varLabels[_itemNumber]);
#endif
	uint8_t itemOffsetInMap = GetOffsetInMap(_itemNumber);
	uint8_t itemOffsetInSequence = GetOffsetInSequence(_itemNumber);
	uint8_t itemBitSize = GetItemBitSize(_itemNumber);
	_mask = 1;
#ifdef _DEBUG_BITFIELD_STORAGE_CREATE_MASK
	Serial.print(F("itemOffsetInSequence = "));
	Serial.print(itemOffsetInSequence);
	Serial.print(F(", itemBitSize = "));
	Serial.print(itemBitSize);
	Serial.print(F(", _mask "));
	Serial.println(_mask, BIN);
#endif

#ifdef _DEBUG_BITFIELD_STORAGE_CREATE_MASK
#endif
	for (int i = 0; i < itemBitSize - 1; i++) {
		_mask = _mask << 1;
		_mask += 1;
	} // for
	_mask = _mask << itemOffsetInSequence;
#ifdef _DEBUG_BITFIELD_STORAGE_CREATE_MASK
	Serial.print(F("_mask initialized to "));
	Serial.println(_mask, BIN);
#endif
	uint8_t indexIntoRegister = GetOffsetInSequence(_itemNumber);
	_mask <<= (indexIntoRegister - 8 + 1);
#ifdef _DEBUG_BITFIELD_STORAGE_CREATE_MASK
	Serial.print(F("Final mask "));
	Serial.println(_mask, BIN);
#endif

#ifdef _DEBUG_BITFIELD_STORAGE_CREATE_MASK
	PRINT_FUNCTION_EXIT_NL
#endif
		return _mask;
} // CreateMask()

// Begin BitfieldStorage.cpp
