#ifndef BITFIELD_STORAGE_H
#define BITFIELD_STORAGE_H

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// incorporate it inside of an executable or other machine language component
// without restriction for personal or commercial use.The Software may not be
// redistributed as source code or in any recognizable human readable form
// in any form whatsoever for any use whatsoever.
// 
// The Software is the property of Michle Delsol(France), copyright 2023 USA and
// international. For any questions concerning use of the software contact
// Michle Delsol at CPParduino@md-dsl.fr.
// 
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include "Globals.h"

// This framework enables access to uint8_t data stored as bitfields to optimize data storage.
// For example, a month has values which range from 1 to 12 (0 to 11), therefore only
// only 4 bits are required 0 to 15
// Day of the month requires 5 bits, 1 to 31 (0 to 31)
// Each value is characterized by a bit size and a calculated offset must
// fill 8 bit sequences completely, via dummy variables if necessary.
// Straddling across 8-bit boundaries is not allowed.

// Classes which use this framework must put their bitfields inside a structure
// Two enums need to be created:
//   enum of variables
//   uint8_t array of bit sizes
//   char* array of variable names
// The above must be synchronized

struct VarCharacteristics {
	uint8_t size;
	uint8_t min, max;
	char* label;
}; // VarCharacteristics


class BitfieldStorage {
  public:
//-------------------------------- CONSTRUCTOR -------------------------------
    BitfieldStorage (uint8_t* _valuesArray, uint8_t _varCount);

//-------------------------------- GET/SET FUNCTIONS -------------------------------
    void Set(uint8_t _itemNumber, uint8_t _valueToSet);
    uint8_t  Get(uint8_t _itemNumber);
// ------------------------------ AUXILIARY PUBLIC FUNCTIONS -------------------------
    void PrintVariables();
    void PrintMasks();
    boolean VerifyMasks();
//-------------------------------- AUXILIARY PRIVATE FUNCTIONS ---------------------------------
    //private:
    uint8_t GetOffsetInMap(uint8_t _itemNumber);
    uint8_t GetSequenceIndex(uint8_t _itemNumber);
    uint8_t GetOffsetInSequence(uint8_t _itemNumber);
    uint8_t GetItemBitSize( uint8_t _itemNumber);
    boolean ValuesArrayCoherence(uint8_t& _faultyIndex);
    uint8_t BitfieldStorage::CreateMask( uint8_t _itemNumber, uint8_t& _mask);
    uint16_t GetDataStorageSize() { return sizeof(valuesArray); };
    uint16_t GetVarCount() { return varCount; };
    //-------------------------------- PRIVATE VARIABLES ---------------------------------
 public:
    static VarCharacteristics varCharacteristics[];
 private:
    uint8_t* valuesArray{(uint8_t*) 0 }; // maps values of bitfields
    uint8_t varCount{0};
 
}; // BitfieldStorage

#endif // BITFIELD_STORAGE_H
