/* ATest.cpp
called by Aardvark to test stuff
*/

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// incorporate it inside of an executable or other machine language component
// without restriction for personal or commercial use.The Software may not be
// redistributed as source code or in any recognizable human readable form
// in any form whatsoever for any use whatsoever.
// 
// The Software is the property of Michle Delsol(France), copyright 2023 USA and
// international. For any questions concerning use of the software contact
// Michle Delsol at CPParduino@md-dsl.fr.
// 
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
// 
// Scenario
// A joblist is a list of jobs. A job is a list of letter digit pairs which define it. 
// Examples:
// D2130 (length 5) - do data everyday indefinitely at 2130
// M30000 (length 6) - do maintenance on day 3 (Wednesday) indefinitely at midnight
// S150000 (length 7) - do sync on the 15th at indefinitely midnight
// D1430R10M30 (length 11) - do data at 1430, repeat 10 times every 30 minutes
// D06031430R5M30V05 (length 18) - do data on June 3rd at 1430, repeat 5 times every 30 mins. 5 days validity
// Fields: job type, timing, repeats, validity
// Job type: 'D' 'M' 'S' (data, synchronize or maintenance)
// Timing format:
// hhmm - everyday at a specific time.
// whhmm - once a week,  w = 1 to 7, 1 being Monday
// mmhhmm - once a month on a specific day of the month at a specifc time, not that the 5th would be 05, not 5
// Repeats:
// RxxMmm - indefinite planned task starts at time, xx repeats, interval mm minutes or hh hours
// Repeats: Start with letter Rxx(M/H)xx
// Empty is no repeats, run once only
// Rxx is number of repeats - R5 is repeat 5 times
// (M/H)xx is interval in min or hours of repeats M15 is repeat every 15 minutes
// daysValidity : V5 - 'V' followed by two ascii chars which define days validity including start date
// 
// A regex could be used to capture the data:
// ([DMS])([0-9]{4-6})([R]|\s)?([0-9]{1-2})([MH]|\s)?(([0-9]{1-2}))([V]|\s)?([0-9]{1-2})
// ^      ^           ^        ^           ^         ^             ^        ^
// job    timing      repeats  # of rep.   min/hr    duration      validity days valid
// Validity checks on captures establish the job's acceptability.
// 
// Once jobs gave been defined, constraints need to be verified to prevent one job clashing with another job.
// There must a 10 minute interval between one job's endTime and another job's startTime. This means that a job
// which does not repeat cannot execute inside the time slot of a job which repeats.
// 
// Beehive weighing system jobs are stored as variable length strings in EEPROM via the DataGroup class
// No attempt is made to validate or parse the jobs, this is up to the Job class to do so, not within the context of this program.
// 
// A DataGroups class is created after which five jobs (variable length, offset access) are added to it.
// Some jobs are deleted. The data group is then defragmented.
// The process shows how the data group takes advantage of holes when new jbs get added.
//

#include "Globals.h"

// Comment/uncomment according to what needs getting done

#define CREATE_NEW_DATAGROUP // ***************************************************************
#define TEST_REMOVE_RECORD
#define TEST_INSERT_RECORD

void ATest() {
	PRINT_FUNCTION_ENTER_NL
	#ifdef ARDUINO_IDE
	Serial.println(F("Using Visual Micro"));
	#else
	Serial.println(F("Using AtmelStudio"));
	#endif

	bool success{false};
	char* record{0};

	#ifdef CREATE_NEW_DATAGROUP

	char* jobsArray[] = {
		"D2130", // (length 5) - do data everyday indefinitely at 2130
		"M30000", // (length 6) - do maintenance on day 3 (Wednesday) indefinitely at midnight
		"S150000", // (length 7) - do sync on the 15th at indefinitely midnight
		"D06031430R5M30V05", // (length 18) - do data on June 3rd at 1430, repeat 5 times every 30 mins. 5 days validity
		"D1430R10M30" // (length 11) - do data at 1430, repeat 10 times every 30 minutes
	};

	// Create a data group for jobs
	DataGroup jobsDataGroup = DataGroup(
	DATA_GROUP_SIZE, // Total allocated bytes
	"BeehiveJobs", // Data group's name
	"Beehive jobs stored as variable length strings in EEPROM", // strlen = 56 bytes - data group's description
	EDataType::charData, ERecordLength::variable, ERecordAcccess::offset // Data group characteristics
	);

	if ( jobsDataGroup.GetDataGroupSize() == 0 ) {
		Serial.println("ATest ERROR - failed to create DataGroup");
		PRINT_FUNCTION_EXIT_NL
		return;
		} else {
		PRINT_VALUE_EQUAL_DESC_NL("ATest, DataGroup created - name: ", jobsDataGroup.GetName());
	}

	// Populate DataGroup with beehive jobs data
	// Define number of items and total array size
	uint16_t arrayCount = sizeof(jobsArray)/sizeof(jobsArray[0]);
	uint16_t arraySize{0};
	for ( int i = 0; i < arrayCount; i++) {
		arraySize += strlen(jobsArray[i]);
	}
	arraySize += arrayCount*2; // Add record size header  to get total requirement

	success = jobsDataGroup.RecordsAdd(jobsArray, arrayCount, EDataType::charData);
	if ( success == false) {
		Serial.println("ATest ERROR - failed to populate data group");
		} else {
		Serial.println("ATest ERROR - success populating data group");
	}

	#else // CREATE_NEW_DATAGROUP - USE EXISTING DATA GROUP*****************************************************************
	// Open existing data group
	// Set to true to initialize data group to zero records
	char* datagroupName{"BeehiveJobs"}; // Should exist
	//char* datagroupName{"Fake"}; // For testing purposes
	DataGroup jobsDataGroup = DataGroup(datagroupName);
	// Check that requested DataGroup exists
	if ( strncmp(jobsDataGroup.GetName(), "WRONG", strlen("WRONG")) == 0  ) {
		PRINT_VALUE_EQUAL_DESC_NL("ATest ERROR - following data group requested does not exist :", datagroupName);
		goto errorLabel;
	}
	#endif // CREATE_NEW_DATAGROUP

	#ifdef TEST_REMOVE_RECORD
	record = "M30000";
	strcpy(record,"M30000"); // (length 6) - do maintenance on day 3 (Wednesday) indefinitely at midnight
	success = jobsDataGroup.RecordRemove(record, EDataType::charData);
	if (success == true) {
		PRINT_VALUE_EQUAL_DESC_NL("Removed: ", record);
		} else {
		PRINT_VALUE_EQUAL_DESC_NL("Remove failed on record: ", record);
	}

	record = "D06031430R5M30V05"; // (length 17) - do maintenance on day 3 (Wednesday) indefinitely at midnight
	success = jobsDataGroup.RecordRemove(record, EDataType::charData);
	if (success == true) {
		PRINT_VALUE_EQUAL_DESC_NL("Removed: ", record);
		} else {
		PRINT_VALUE_EQUAL_DESC_NL("Remove failed on record: ", record);
	}
	#endif // TEST_REMOVE_RECORD  *********************************************

	#ifdef TEST_INSERT_RECORD //***********************************************
	record = "Longer13bytes"; // (length 13) - do data on June 3rd at 1430, repeat 5 times every 30 mins. 5 days validity
	success = jobsDataGroup.RecordInsert(jobsDataGroup.GetHeaderSize(), record);
	if (success == true) {
	PRINT_VALUE_EQUAL_DESC_NL("Inserted: ", record);
	} else {
	PRINT_VALUE_EQUAL_DESC_NL("Insertion failed on record: ", record);
	}

	record = "5byte"; // (length 5) - do data on June 3rd at 1430, repeat 5 times every 30 mins. 5 days validity
	success = jobsDataGroup.RecordInsert(jobsDataGroup.GetHeaderSize(), record);
	if (success == true) {
	PRINT_VALUE_EQUAL_DESC_NL("Inserted: ", record);
	} else {
	PRINT_VALUE_EQUAL_DESC_NL("Insertion failed on record: ", record);
	}

	record = "VeryLongRecord21bytes"; // (length 21) - do data on June 3rd at 1430, repeat 5 times every 30 mins. 5 days validity
	success = jobsDataGroup.RecordInsert(jobsDataGroup.GetHeaderSize(), record);
	if (success == true) {
	PRINT_VALUE_EQUAL_DESC_NL("Inserted: ", record);
	} else {
	PRINT_VALUE_EQUAL_DESC_NL("Insertion failed on record: ", record);
	}
	#endif // TEST_INSERT_RECORD  ********************

	jobsDataGroup.RecordsPrint("Before defragment");
	bool foundHole = jobsDataGroup.Defragment();
	Serial.println("Defragment found hole");
	jobsDataGroup.RecordsPrint("After defragment");
	PRINT_FUNCTION_EXIT_NL
	return;
	//Exit via here if there are errors
	errorLabel:
	PRINT_VALUE_EQUAL_DESC_NL("ATest exit with error: ", "Need implement error reporting.");
	PRINT_FUNCTION_EXIT_NL
	return;
	} // ATest()

	// End of ATest.cpp
