#ifndef DATA_GROUP_H
#define DATA_GROUP_H

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// incorporate it inside of an executable or other machine language component
// without restriction for personal or commercial use.The Software may not be
// redistributed as source code or in any recognizable human readable form
// in any form whatsoever for any use whatsoever.
// 
// The Software is the property of Michle Delsol(France), copyright 2023 USA and
// international. For any questions concerning use of the software contact
// Michle Delsol at CPParduino@md-dsl.fr.
// 
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include "Globals.h"

// DataGroup class stores fixed and variable length records inside a continuous memory spce
// There is a header which contains various info
// Data is a succession of records:
// For variable length records,
//		first two bytes (uint16_t) define record size
//		next bytes contain data - first byt equal 0 signifies empty slot
// For fixed length records, they are sized according to their
//		corresponding array size - empty slot is space filled with 0xFF values.
// Access can be sequential or key based - currently sequential only.

// Current implementation is EEPROM based storage. Future development,
//		storage should be transparent, a #define should suffice to determine where to store.
// When removing a record, a hole gets created. It can be identified
//		by its first record position being 0, thus no records which
//		start with 0 are allowed

#include <avr/eeprom.h>
// for read/write to EEPROM use :
// void eeprom_write_block (const void *__src, void *__dst, size_t __n);
// void eeprom_read_block (void *__dst, const void *__src, size_t __n);

#if defined(ATMEGA_2560)
#define MAX_MEMORY_ALLOC 4096
#elif defined(ATMEGA_328P)
#define MAX_MEMORY_ALLOC 1024
#else
#define MAX_MEMORY_ALLOC 0
#error ("ERROR - should define ATMEGA_328P or ATMEGA_2560")
#endif

#define DATA_GROUP_SIZE 350 // Use define when creating a new data group
#if DATA_GROUP_SIZE > MAX_MEMORY_ALLOC && defined(ATMEGA_2560)
#error ("DataGroup size should be less than EEPROM size : 4096")
#elif DATA_GROUP_SIZE > MAX_MEMORY_ALLOC && defined(ATMEGA_328P)
#error ("DataGroup size should be less than EEPROM size : 1024")
#else
#endif

// Data group parameters
#define DATA_GROUP_NAME_LENGTH 20
#define DATA_GROUP_DESCRIPTION_LENGTH 100
// Safety measure to avoid runaway situation
// Also helps detect programming error
#define MAX_COUNT 16 

#define _USE_EEPROM
#define _USE_RAM
#if defined(_USE_EEPROM)
// Not generalized - used in long DataGroup constructor once to verify
#define WRITE_RECORD(_DATA, _OFFSET, _SIZE,_OFFSET_ADJUST) /* EEPROM */	\
eeprom_write_block(_DATA, _OFFSET, _SIZE); \
_OFFSET = _OFFSET + _OFFSET_ADJUST;
#define WRITE_BYTE(_DATA, _OFFSET, _SIZE,_OFFSET_ADJUST) /* EEPROM */	\
eeprom_write_byte(_OFFSET, _DATA); 
#elif defined(_USE_RAM)
#error "ERROR - RAM memory storage not yet implemented"
#endif // _USE_EEPROM

enum class EDataType : uint8_t {  charData, int8Data, int16Data, floatData };
enum class ERecordLength : uint8_t { fixed, variable };
enum class ERecordAcccess : uint8_t { offset, key };

	struct StrDGheader  {
		char name[DATA_GROUP_NAME_LENGTH] {""}; // Includes end null
		char desc[DATA_GROUP_DESCRIPTION_LENGTH] {""}; // Includes end null
		uint16_t dataGroupSize{0}; // defined in constructor
		uint16_t recordSize{0}; // 0 for variable length otherwise > 0 for fixed length records
		uint16_t itemsCount{0};
		// Defaults for variable length char data
		EDataType dataType{EDataType::charData};
		ERecordLength dataLength{ERecordLength::variable};
		ERecordAcccess dataAccess{ERecordAcccess::offset};
	};

class DataGroup {
	public:
	// Constructors
	DataGroup(char* _name);  // Open/verify data group
	DataGroup(uint16_t _dataGroupSize, char* _name, char* _desc, EDataType _dataType, ERecordLength _type, ERecordAcccess _access); // Create data group
	// Methods
	bool RecordsAdd(char** _data, uint16_t _arrayCount, EDataType _dataType);
	void RecordsPrint(char* _desc); // Prints all records
	bool RecordInsert( uint16_t _startOffset, char* _dataToInsert ); //Finds first empty slot as of _startOffset
	bool RecordRemove(char* _data, EDataType _dataType);
	uint16_t RecordGet(uint16_t _offset, char* _segment); // Finds record based on content
	bool Defragment(); // Remove holes
	void HeaderDataPrint(char* _desc); // Prints data group info
	void HeaderDataRead(); // Reads data group header info from EEPROM
	void HeaderDataWrite(); // Writes data group header info from EEPROM
	// Get variables
	char* GetName() { return strDGheader.name;}
	uint16_t GetDataGroupSize() { return strDGheader.dataGroupSize;}
	uint16_t GetHeaderSize(){ return sizeof(strDGheader);}
	uint16_t GetRecordCount(){ return strDGheader.itemsCount;}
	// Records variables in databse
	void RecordItemsCountInEEPROM();
	uint16_t GetLastAvailableOffset(); // Gets space after last record

	private:
	// Data
	StrDGheader strDGheader;
	const uint16_t remainingSpace{strDGheader.dataGroupSize - sizeof(strDGheader)};
	// TODO
	uint16_t GetRecord(uint16_t* _subArrayToFind, uint16_t _subArrayToFindLength);
}; // class EEPROMstore

#endif // DATA_GROUP_H
