#ifndef DATA_PACKETS_H
#define DATA_PACKETS_H

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// incorporate it inside of an executable or other machine language component
// without restriction for personal or commercial use.The Software may not be
// redistributed as source code or in any recognizable human readable form
// in any form whatsoever for any use whatsoever.
// 
// The Software is the property of Michle Delsol(France), copyright 2023 USA and
// international. For any questions concerning use of the software contact
// Michle Delsol at CPParduino@md-dsl.fr.
// 
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

// PROGRAM CLASS START
// The DataPackets class transforms z chunk of data into packets to be
// sent / retrieved via an externally defined media, be it a serial port,
// Wifi, or whatever.Two methods handle the process :
// GeneratePacketsArray - receives a chunk od data, transforms it
// into a packet array.
// GenerateDataChunk - called repeatedly to get packets od data to
// assemble them into a chunk of data.
// 
// Data gets sent as groups of packets - 4 packets max,
// Total packet size is 48 
//    First packet can hold 48 - 8 = 40 bytes of data
//    Next packets can hold 48 - 6 = 42 bytes of data
//    Total  40 + 3*42 = 166 bytes
// |------head bytes----|------------data size-----------------|---end bytes---|
// |--------------------------------packet size--------------------------------|

// Packet structure index : byte contents
// 0 Start code
// 1 packet no. - 1 to 4
// 2 packetSize 
// 3 DataID - user defined
// 4 packetCount (first packet only)
// 5 chunk size (169 max) (first packet only)
// 6 begin data
// packetSize-3 : end data
// packetSize-2 : Checksum
// packetSize-1 : End code
// PROGRAM FILE SET END DataPackets


//---------------------------- TRANSFER BASIC PARAMETERS -----------------------------------------------
// data size 
#define DATA_PACKETS_MAX_COUNT 4 // max no of packets
#define DATA_PACKETS_MAX_PACKET_SIZE 48 // max total packet size
#define OVERHEAD_FIRST_PACKET 8
#define OVERHEAD_NEXT_PACKET 6
//#define DATA_PACKETS_MIN_PACKET_SIZE 16 // min total packet size
#define DATA_SIZE_FIRST_PACKET (DATA_PACKETS_MAX_PACKET_SIZE - OVERHEAD_FIRST_PACKET)
#define DATA_SIZE_NEXT_PACKETS (DATA_PACKETS_MAX_PACKET_SIZE - OVERHEAD_NEXT_PACKET)
#define DATA_CHUNK_MAX_SIZE (DATA_SIZE_FIRST_PACKET+3*DATA_SIZE_NEXT_PACKETS)

// OVERALL DATA OFFSETS when transmitting: header - data - tails
#define DATA_PACKETS_START_CODE 0x7E
#define DATA_PACKETS_END_CODE 0x7F
#define DATA_PACKETS_LENGTH_START_BYTE               1 // length of start byte
#define DATA_PACKETS_LENGTH_END_BYTE                 1 // length of end byte

// Global functions
bool VerifyParametersCoherenceGlobal(uint8_t* _packetsPtr, uint8_t _packetCount, uint8_t _chunkSize, uint8_t _packetID, uint8_t _noItemsToVerify, uint8_t _packetSize);
void ListPacketContents(uint8_t* _packetPtr, uint8_t _packetID);

class DataPackets {
	// generic class which handles basic send/receive XBee data
public:
	// Send data constructor
	DataPackets();
	// Get data constructor
	bool GeneratePacketsArray(uint8_t* _chunkData, uint16_t _chunkSize, uint8_t _dataID);
	bool GetDataFromPacketsArray(uint8_t** _packetsArray, uint8_t* _destinationArray);
	uint8_t GetDataFromPacket(uint8_t* _packetData, uint8_t _packetID);
	uint8_t* GetPacket(uint8_t _packetID) { return packetsArray[_packetID-1]; }
	uint8_t* GetChunk() { return chunkData; }
	uint8_t GetNoPacketsCount() { return noPacketsToSendReceive; }
	bool VerifyParametersCoherence(uint8_t _packetID, uint8_t _noItemsToVerify, uint8_t _packetSize);
	uint8_t GetPacketSizeForPacketID(uint8_t _packetID, uint8_t _totalNoPackets, uint16_t _chunkSize); // Gets how many bytes to send for each packet
private: // functions
	uint8_t GetDataBufferArraySize() { return chunkSize; }
	uint8_t CalculateChecksum(byte* _dataBuffer, uint8_t _start, uint8_t _end);
	bool VerifyChecksum(uint8_t* _packet, uint8_t _packetSize, uint8_t _start, uint8_t _end);
	uint8_t CalculateNoPacketsFromChunkSize(uint16_t _dataSize);
	void CheckPacketDataValidity(uint8_t _packetID, uint8_t _startInPacketOffset, uint16_t _startInChunkOffset, uint8_t _compareLength);
	void ListPacketData(uint8_t _packetID, uint8_t _packetSize, bool _listAsChar);
private: // variables
	uint8_t* chunkData{ 0 }; // Data to be sent/received
	uint16_t chunkSize{ 0 }; // total bytes in data to be sent/received
	uint8_t chunkID{ 0 }; // ID of data to be sent/received
	uint8_t noPacketsToSendReceive{ 1 }; // No. of packets chunkData converts to
	uint16_t packetSize{ 0 }; // total bytes in data to be sent/received
	uint8_t* packetsArray[DATA_PACKETS_MAX_COUNT]; // Packets array chunkData converts to
}; // class DataPackets

#endif // DATA_PACKETS_H
