// Begin EventStorage.cpp

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// incorporate it inside of an executable or other machine language component
// without restriction for personal or commercial use.The Software may not be
// redistributed as source code or in any recognizable human readable form
// in any form whatsoever for any use whatsoever.
// 
// The Software is the property of Michle Delsol(France), copyright 2023 USA and
// international. For any questions concerning use of the software contact
// Michle Delsol at CPParduino@md-dsl.fr.
// 
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include "Globals.h"

//------------------------- BITFIELD STORAGE -------------------------------
#define _DEBUG_EVENT_STORAGE_ON
#ifdef _DEBUG_EVENT_STORAGE_ON

#define _DEBUG_EVENT_STORAGE_EVENT_STORAGE
//#define _DEBUG_EVENT_STORAGE_GET
//#define _DEBUG_EVENT_STORAGE_SET
#define _DEBUG_EVENT_STORAGE_VERIFY_MASKS
#define _DEBUG_EVENT_STORAGE_PRINT_VARIABLES
#define _DEBUG_EVENT_STORAGE_PRINT_MASKS
#define _DEBUG_EVENT_STORAGE_PRINT_RAW_DATA
//#define _DEBUG_EVENT_STORAGE_CREATE_MASK
//#define _DEBUG_EVENT_STORAGE_GET_OFFSET_IN_MAP

#endif // _DEBUG_EVENT_STORAGE_ON

EventStorage::EventStorage(uint8_t _registerSize, uint8_t _eventCount, char* _eventLabels[])
	: registerSize(_registerSize), eventCount(_eventCount), eventLabels(_eventLabels) {
	// Allocates space for event bytes according to event count and register size
	// Intializes event storage array ietms to zero (i.e. not triggered)
	// Constructor. Checks storage array coherence via ValuesArrayCoherence()
	// Parameters
	//uint8_t* _eventsArray - pointer to the first variable in the list of variables
	//uint8_t* _registerSize - 7 or 8 bit registers
	//uint8_t _eventCount - the total number of variables
	//char* _eventLabels[] - labels to identify the variables for reporting

#ifdef _DEBUG_EVENT_STORAGE_EVENT_STORAGE
	PRINT_FUNCTION_ENTER_NL;
#endif

#ifdef _DEBUG_EVENT_STORAGE_EVENT_STORAGE
	PRINT_VALUE_EQUAL_TAB_NL(_registerSize);
	PRINT_VALUE_EQUAL_TAB_NL(_eventCount);
#endif

	// Allocate space and initialize events array
	eventsArray = new uint8_t(eventCount / registerSize + eventCount % registerSize ? 1 : 0);

#ifdef _DEBUG_EVENT_STORAGE_EVENT_STORAGE
	PRINT_FUNCTION_ENTER_NL;
#endif
}; // EventStorage constructor

bool EventStorage::Reset() {
	for (int i = 0; i < eventCount; i++) {
		Set(i, 0);
	}
}

//-------------------------------- GET/SET FUNCTIONS -------------------------------
bool EventStorage::Get(uint8_t _itemNumber, uint8_t& _valueToGet) {
#ifdef _DEBUG_EVENT_STORAGE_GET
	PRINT_FUNCTION_ENTER_NL;
	Serial.print(F("Getting value for _itemNumber =  "));
	Serial.print(_itemNumber);
	Serial.print(F("  "));
	Serial.println(eventLabels[_itemNumber]);
#endif
	int8_t regIndex = GetSequenceIndex(_itemNumber);
	uint8_t offsetInMap = GetOffsetInMap(_itemNumber);
	uint8_t offsetInSequence = GetOffsetInSequence(_itemNumber);

	uint8_t offsetEnd{ offsetInSequence + 1 - 1 };
	//uint8_t indexInRegister{ offsetInMap - (8 * regIndex)};

#ifdef _DEBUG_EVENT_STORAGE_GET
	Serial.print(F("regIndex = ")); Serial.println(regIndex);
	Serial.print(F("bitSize = ")); Serial.println(1);
	Serial.print(F("offsetInSequence = ")); Serial.println(offsetInSequence);
	Serial.print(F("offsetEnd = ")); Serial.println(offsetEnd);
	//Serial.print(F("indexInRegister = ")); Serial.println(indexInRegister);
#endif

  // From here on all parameters should be OK
	uint8_t mask{ 0 };
	CreateMask(_itemNumber, mask);
#ifdef _DEBUG_EVENT_STORAGE_GET
	PRINT_VALUE_EQUAL_BIN_NL(mask);
#endif

	// Now do work
	// Simulate one byte at offset into the mapped values
	uint8_t* mappedBytes = eventsArray;
	//uint8_t* mappedBytes = (uint8_t * )this; // Replace with data from 

	uint8_t tempValue{ 0 };
	tempValue = mappedBytes[regIndex];
#ifdef _DEBUG_EVENT_STORAGE_GET
	Serial.print(F("tempValue ")); Serial.println(tempValue, BIN);
#endif
	//Serial.print(F("Just before applying mask tempValue = ")); Serial.println(tempValue);
	tempValue &= mask;
	//Serial.print(F("Just after applying mask tempValue = ")); Serial.println(tempValue);
	tempValue >>= (offsetInSequence - 1 + 1);
	_valueToGet = tempValue;

#ifdef _DEBUG_EVENT_STORAGE_GET
	PRINT_FUNCTION_EXIT_NL;
#endif
	return true;
} // Get()

bool EventStorage::Set(uint8_t _itemNumber, uint8_t _valueToSet) {
#ifdef _DEBUG_EVENT_STORAGE_SET_EVENT_VALUE
	PRINT_FUNCTION_ENTER_NL;
	Serial.print(F("Entering EventStorage::")); Serial.println(__FUNCTION__);
	Serial.print(F("Setting value for _itemNumber =  "));
	Serial.print(_itemNumber);
	Serial.print(F("  "));
	Serial.print(eventLabels[_itemNumber]);
	Serial.print(F(", _valueToSet = "));
	Serial.println(_valueToSet);
#endif
	int8_t regIndex = GetSequenceIndex(_itemNumber);
	uint8_t offsetInMap = GetOffsetInMap(_itemNumber);
	uint8_t offsetInSequence = GetOffsetInSequence(_itemNumber);
	uint8_t bitSize = 1;
	uint8_t offsetEnd{ offsetInSequence + bitSize - 1 };
	//uint8_t indexInRegister{offsetInMap - (8 * regIndex)};

#ifdef _DEBUG_EVENT_STORAGE_SET_EVENT_VALUE
	Serial.print(F("Entering EventStorage::")); Serial.println(__FUNCTION__);
	Serial.print(F("regIndex = ")); Serial.println(regIndex);
	Serial.print(F("offsetInMap = ")); Serial.println(offsetInMap);
	Serial.print(F("offsetInSequence = ")); Serial.println(offsetInSequence);
	Serial.print(F("bitSize = ")); Serial.println(bitSize);
	Serial.print(F("offsetEnd = ")); Serial.println(offsetEnd);
	//Serial.print(F("indexInRegister = ")); Serial.println(indexInRegister);
#endif

  // From here on all parameters should be OK
	uint8_t mask{ 0 };
	CreateMask(_itemNumber, mask);
#ifdef _DEBUG_EVENT_STORAGE_SET_EVENT_VALUE
	Serial.print(F("\tmask ")); Serial.println(mask, BIN);
#endif

	// Now do work
	// Simulate one byte at offset into the mapped values
	uint8_t* mappedBytes = eventsArray;
	//uint8_t* mappedBytes = (uint8_t*)this; // does not work
	// Down to here Get and Set are identical

	uint8_t tempValue{ _valueToSet };
	tempValue <<= (offsetInSequence - bitSize + 1);
	mappedBytes[regIndex] &= ~mask;
	mappedBytes[regIndex] += tempValue;

#ifdef _DEBUG_EVENT_STORAGE_SET_EVENT_VALUE
	PRINT_FUNCTION_EXIT_NL;
	Serial.println(F("After processing "));
	Serial.print(F("Address eventsArray ")); Serial.println((long)eventsArray);
	Serial.print(F("Address mappedBytes ")); Serial.println((long)mappedBytes);
	Serial.print(F("regIndex ")); Serial.print(regIndex);
	Serial.print(F(" value content ")); Serial.print(mappedBytes[regIndex]);
	Serial.print(F(" bin content ")); Serial.println(mappedBytes[regIndex], BIN);
#endif
	return true;
} // Set()

// ------------------------------ AUXILIARY PUBLIC FUNCTIONS -------------------------
bool EventStorage::VerifyMasks() {
#ifdef _DEBUG_EVENT_STORAGE_VERIFY_MASKS
	PRINT_FUNCTION_ENTER_NL;
#endif

	uint8_t totalSequenceMask{ 0 };
	uint8_t sequenceCount = (GetOffsetInMap(eventCount - 1) + 1) / 8;
	uint8_t sequence{ 0 };
	for (int i = 0; i <= eventCount; i++) {
		uint8_t offsetInSequence = GetOffsetInSequence(i);
		if (offsetInSequence == 0) {
			if (i > 0) {
				if (totalSequenceMask != 255) {
					Serial.print(F("ERROR before item ")); Serial.print(i);
					Serial.print(F(" ")); Serial.print(eventLabels[i]);
				} // if
				Serial.print(F("TotalMaskSequence = ")); Serial.println(totalSequenceMask, BIN);
			}
			if (i < eventCount) {
				Serial.print(F("sequence = ")); Serial.println(sequence);
			}
			if (i == eventCount) break;
			totalSequenceMask = 0;
			sequence++;
		}
		uint8_t mask{ 0 };
		mask = CreateMask(i, mask);
		totalSequenceMask += mask;
		Serial.print(eventLabels[i]);
		Serial.print(F(" mask = "));
		Serial.print(mask, BIN);
		Serial.print(F(" maskSequence  = ")); Serial.println(totalSequenceMask, BIN);
	}// for

#ifdef _DEBUG_EVENT_STORAGE_PRINT_MASKS
	PRINT_FUNCTION_EXIT_NL;
#endif
	return true;
} // VerifyMasks()

void EventStorage::PrintMasks() {
#ifdef _DEBUG_EVENT_STORAGE_PRINT_MASKS
	PRINT_FUNCTION_ENTER_NL;
#endif
	for (int i = 0; i < eventCount; i++) {
		uint8_t mask{ 0 };
		mask = CreateMask(i, mask);
		uint8_t offsetInSequence = GetOffsetInSequence(i);
		uint8_t sequence{ 0 };
		if (offsetInSequence == 0) {
			Serial.print(F("sequence = "));
			Serial.println(sequence++);
		}
		Serial.print(F("i = "));
		Serial.print(i);
		Serial.print(F("\t"));
		Serial.print(eventLabels[i]);
		Serial.print(F(" offsetInSequence = "));
		Serial.print(offsetInSequence);
		Serial.print(F(" bitSize = "));
		Serial.print(1);
		Serial.print(F(" mask = "));
		Serial.println(mask, BIN);
	} // for

#ifdef _DEBUG_EVENT_STORAGE_PRINT_MASKS
	PRINT_FUNCTION_EXIT_NL;
#endif
	return;
} // PrintMasks()

void EventStorage::PrintRawData() {
#ifdef _DEBUG_EVENT_STORAGE_PRINT_RAW_DATA
	PRINT_FUNCTION_ENTER_NL;

	for (int i = 0; i < eventCount / registerSize + 1; i++) {
		Serial.print(F("eventsArray[i] ")); Serial.println(eventsArray[i], BIN);
	}

	Serial.print(F("Exiting EventStorage::")); Serial.println(__FUNCTION__);
	PRINT_FUNCTION_EXIT_NL;
#endif
}// PrintRawData()

void EventStorage::PrintVariables() {
#ifdef _DEBUG_EVENT_STORAGE_PRINT_VARIABLES
	PRINT_FUNCTION_ENTER_NL;
	Serial.print(F("eventCount = ")); Serial.println(eventCount);
#endif
	for (int i = 0; i < eventCount; i++) {
		uint8_t itemValue{ 0 };
		uint8_t sequenceIndex = GetSequenceIndex(i);
		uint8_t offsetInMap = GetOffsetInMap(i);
		Get(i, itemValue);
		if (offsetInMap % 8 == 0) {
#ifdef _DEBUG_EVENT_STORAGE_PRINT_VARIABLES
			Serial.print(F("Sequence = ")); Serial.println(sequenceIndex);
#endif
		}
		Serial.print(F("\ti = "));
		Serial.print(i);
		Serial.print(F("\t"));
		Serial.print(eventLabels[i]);
		Serial.print(F(" = "));
		Serial.println(itemValue);
	} // for
#ifdef _DEBUG_EVENT_STORAGE_PRINT_VARIABLES
	PRINT_FUNCTION_EXIT_NL;
#endif
	return;
} // PrintVariables()

//-------------------------------- AUXILIARY PRIVATE FUNCTIONS ---------------------------------
uint8_t EventStorage::GetOffsetInMap(uint8_t _itemNumber) {
#ifdef _DEBUG_EVENT_STORAGE_GET_OFFSET_IN_MAP
	PRINT_FUNCTION_ENTER_NL;
#endif
	if (_itemNumber == 0) return 0;
	uint8_t offset{ 0 };
	for (int i = 1; i <= _itemNumber; i++) {
		offset += 1;
		continue;
#ifdef _DEBUG_EVENT_STORAGE_GET_OFFSET_IN_MAP
		Serial.print(F("i = "));
		Serial.print(i);
		Serial.print(F(", bitSize i-1 = "));
		Serial.print(varSizes[i - 1]);
		Serial.print(F(", offset = \t"));
		Serial.println(offset);
#endif
	}
#ifdef _DEBUG_EVENT_STORAGE_GET_OFFSET_IN_MAP
	PRINT_FUNCTION_EXIT_NL;
#endif
	return offset;
}

uint8_t EventStorage::GetSequenceIndex(uint8_t _itemNumber) {
	return GetOffsetInMap(_itemNumber) / 8;
}

uint8_t EventStorage::GetOffsetInSequence(uint8_t _itemNumber) {
	return GetOffsetInMap(_itemNumber) % 8;
}

uint8_t EventStorage::CreateMask(uint8_t _itemNumber, uint8_t& _mask) {
	// Creates a mask of type 00011000, position = 3, maskLength = 2
	// When printing get 11000
	// returns mask
#ifdef _DEBUG_EVENT_STORAGE_CREATE_MASK
	PRINT_FUNCTION_ENTER_NL;
	Serial.print(F("Creating mask for _itemNumber =  "));
	Serial.print(_itemNumber);
	Serial.print(F("  "));
	Serial.println(eventLabels[_itemNumber]);
#endif
	uint8_t itemOffsetInMap = GetOffsetInMap(_itemNumber);
	uint8_t itemOffsetInSequence = GetOffsetInSequence(_itemNumber);
	uint8_t itemBitSize = 1;
	_mask = 1;
#ifdef _DEBUG_EVENT_STORAGE_CREATE_MASK
	Serial.print(F("itemOffsetInSequence = "));
	Serial.print(itemOffsetInSequence);
	Serial.print(F(", itemBitSize = "));
	Serial.print(itemBitSize);
	Serial.print(F(", _mask "));
	Serial.println(_mask, BIN);
#endif

#ifdef _DEBUG_EVENT_STORAGE_CREATE_MASK
#endif
	_mask = 1;
	_mask <<= itemOffsetInSequence;
#ifdef _DEBUG_EVENT_STORAGE_CREATE_MASK
	Serial.print(F("_mask = "));
	Serial.println(_mask, BIN);
#endif

#ifdef _DEBUG_EVENT_STORAGE_CREATE_MASK
	Serial.print(F("Returning with mask = ")); Serial.println(_mask, BIN);
	PRINT_FUNCTION_EXIT_NL;
#endif
	return _mask;
} // CreateMask()

// End EventStorage.cpp

