/* Test.cpp
   called by Aardvark to test stuff
*/

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// incorporate it inside of an executable or other machine language component
// without restriction for personal or commercial use.The Software may not be
// redistributed as source code or in any recognizable human readable form
// in any form whatsoever for any use whatsoever.
// 
// The Software is the property of Michle Delsol(France), copyright 2023 USA and
// international. For any questions concerning use of the software contact
// Michle Delsol at CPParduino@md-dsl.fr.
// 
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include "Globals.h"

// Scenario - This project illustrates using two frameworks: 
//		Project Files Framework - organize project files
//		Print-based Debugging Framework - control what gets printed
// 
// Project Files Framework
// When things start getting complicated, the simple .ino setup/loop 
// model will not do. My beehive weighing system project, at 35 files,
// 15000 lines of code, required my being organized. As the project grew,
// I progressively developed a Project Files Framework. It is structured
// around nine file sets, each of which responds to a specific need independent
// of the project's nature:
// 
// Application entry files which contain setup and loop:.ino file(Arduino),
// Sketch.cpp(AtmelStudio and Visual Studio + Visual Micro), main.cpp(PlatformIO).
// 
// Aardvark.h/.cpp - setup calls Aardvark(); loop is not used - this renders 
//		the application easily portable from one IDE to another.
// Test.h/.cpp - When testing code, do so in Test().It is called from Aardvark().
//		It enables undertaking tests without polluting the contents of Aardvark().
// Globals.h/.cpp - Application wide variables and functions. 
// Macros.h - Store macros separately, do not put them in Globals.h.
// BasicClass.h/.cpp - Each class group should be in a dedicated.h/.cpp file pair.
// HelperFunctions.h/.cpp - Low - level general purpose non - application specific.
// Functions.h/.cpp - High - level functions project specific functions.
// FunctionsSKL.h/.cpp - High - level functions sub - subproject specific functions.
// 
// Print-based Debugging Framework
// There two complementary debugging techniques: Hardware-based which depends on 
// specialized hardware (amply covered in my book Defensive C++ Arduino Programming)
// and print-based which relies on printing info from diverse places in the program.
// This second method is familiar to all programmers since they have as of their first 
// programming endeavors used this technique to figure out what their program was doing.
// The problem with this technique is that after a while, the program gets littered with 
// commented out or worse, not commented out unwanted print commands. This is what happened to me
// as I developed my beehive weighing system. As I progressed with development, I managed
// to control the prints proliferation by creating macros to organize debugging.
// 
// The first set ot macros where PRINT_FUNCTION_ENTER and its exit equivalent to signal
// entering/leaving a function. But since I did not want to systematically comment/uncomment
// these macros as I progressed, I created function level macros to control these via #ifdefs.
// The function level macros also served to control macros inside a function. Function  
// macros could thereby be turned on and off by a simple comment/uncomment of the corresponding
// function level macro.
// In very much the same line of thought, all function level macros could be turned on and off by
// a file level macro located in Globals.h. For example:
// The #ifdef _DEBUG_TEST ... #endif controls all function level macros inside the Test.cpp file. 
// This macro is defined in Globals.h.
// Since only one function is defined in Test.cpp, there is only one function level debugging macro, 
// namely _DEBUG_TEST_TEST macro which controls what gets done inside the Test() function.
// These debugging macros have three parts:
//		_DEBUG_ to signal that this is a debugging macro
//		_TEST which defines the file.
//		_TEST again which is the name of the function.
// In the function Test() below you will see that entry/exit macros will get activated only if the
// corresponding function and file level macros are active
// 
// You will find in Globals.h other application wide debugging macros such as GENERAL_REPORTING,
// ERROR_REPORTNG, and more. They enable activating certain types of macros depending on need. 
// And if you want to shut off all reporting, activate ABSOLUTELY_NO_PRINTS, it will undefine all
// global level macros.
// 
 

#ifdef _DEBUG_TEST
#define _DEBUG_TEST_TEST
#endif // _DEBUG_TEST

void ATest() {
#ifdef _DEBUG_TEST_TEST
	PRINT_FUNCTION_ENTER_NL;
#endif // _DEBUG_TEST_TEST

	Serial.println(F("*************** Test() function - Start doing something. ***************"));
	SampleHelperFunction();
	SampleFunctionSKL();
	SampleFunction();
	CallBasicClass();
	Serial.println(F("*************** Test() function - Finished doing stuff. ***************"));


#ifdef _DEBUG_TEST_TEST
	PRINT_FUNCTION_EXIT_NL;
#endif // _DEBUG_TEST_TEST
	return;
} // ATest

// END FILE Test.cpp
