/* ATest.cpp
   called by Aardvark to test stuff
*/

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// incorporate it inside of an executable or other machine language component
// without restriction for personal or commercial use.The Software may not be
// redistributed as source code or in any recognizable human readable form
// in any form whatsoever for any use whatsoever.
// 
// The Software is the property of Michle Delsol(France), copyright 2023 USA and
// international. For any questions concerning use of the software contact
// Michle Delsol at CPParduino@md-dsl.fr.
// 
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include "Globals.h"

// Scenario:
// Michele's hives - When adding a hive to a yard, the system needs to register it. 
// The new hive will be attributed the jobs other hives have. When the new hive gets turned
// on according to the jobs schedule, it will send its ID and sensor data if the job is
// a data mode job, or its ID and name if the job is a maintenance job.
// The coordinator will receive from the new hive either sensor data and a hive ID or the hive's name
// and its ID. If the new hive's first radio transmission is its ID and sensor data, the 
// coordinator will have to register the new hive with a temporary name linked to its ID. 
// When maintenance mode transmission finally occurs, the coordinator will replace the new hive's
// temporary name with its real name. All this fails if the bee keeper is trying to add 
// more hives than the maximum (16 in this example).
// 
// The process is illustrated in the flow diagram image: HivesRegisterNewModule.jpg
// 
// The algorithm is based on a structure (StructPath) which contains data which define 
// the conditions being tested. The set of conditions is represented by the enum EStructPaths, 
// expected outcome is represented by the enum EStructExpectedResults.
// The algorithm is supposed to reflect the flow diagram (HivesRegisterNewModule.jpg). All six
// paths are tested, results being compared with expected outcome. All results should match 
// expected outcome so that the algorithm be declared correct.
//

#define MODULES_LIMIT 16

void ATest() {
	PRINT_FUNCTION_ENTER_NL

	Serial.println(F("Start simulation register module"));
	// Controlling variables
	const int numberModulesLimit{ MODULES_LIMIT };
	EStructExpectedResults outcome = EStructExpectedResults::doNothing;
	for (int i = 0; i < (uint8_t)EStructPaths::end; i++) {
		uint8_t dataMode = (uint8_t)(structPaths[i].eDataMode);
		Serial.print(F("Simulating PATH "));
		Serial.print(i + 1);
		Serial.print(F(" - "));
		Serial.print(eStructPathsLabels[i]);
		Serial.println(F(" ****************************"));
		bool isRegistered = structPaths[i].isRegistered;
		if (isRegistered == false) { // outer if
			if ((uint8_t)structPaths[i].modulesInYard >= numberModulesLimit) { // inner if
				outcome = EStructExpectedResults::maxNumberReached;
			} else {
				switch (dataMode) {
				case (uint8_t)EDataMode::data: // PATH 5
					outcome = EStructExpectedResults::setTempNameAndRegister;
					break;
				case (uint8_t)EDataMode::maintenance: // PATH 6
					outcome = EStructExpectedResults::setRealNameAndRegister;
					break;
				default: // PATH 4
					outcome = EStructExpectedResults::errorBadMode;
					break;
				} // switch 
			} // if (isRegistered == false)
		}
		else { // Module is registered
			// PATH 1
			if (dataMode == (uint8_t)EDataMode::maintenance && strncmp(structPaths[i].name, "Registered", 10) == 0) {
				outcome = EStructExpectedResults::registeredSetRealName;
			}
			else {  // PATH 2
				outcome = EStructExpectedResults::doNothing;
			} // inner if
		} // outer if
		if (outcome == structPaths[i].eStructExpectedResults) {
			Serial.print(F("\tPATH "));
			Serial.print(i + 1);
			Serial.print(F(" successful: "));
			Serial.print(eStructResultsLabels[(uint8_t)structPaths[i].eStructExpectedResults]);
			Serial.print(F(" - outcome: "));
			Serial.println(eStructResultsLabels[(uint8_t)outcome]);
		}
		else {
			Serial.print(F("\tPATH "));
			Serial.print(i + 1);
			Serial.print(F(" failed expected: "));
			Serial.print(eStructResultsLabels[i]);
			Serial.print(F(" - outcome: "));
			Serial.println(eStructResultsLabels[(uint8_t)outcome]);
		}
	} // for (int i = 0; i < (uint8_t)EStructPaths::end; i++) 
	Serial.println(F("End simulation register module"));

	PRINT_FUNCTION_EXIT_NL	
} // ATest()



// End ATest.cpp
