/* ATest.cpp
   called by Aardvark to test stuff
*/

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// incorporate it inside of an executable or other machine language component
// without restriction for personal or commercial use.The Software may not be
// redistributed as source code or in any recognizable human readable form
// in any form whatsoever for any use whatsoever.
// 
// The Software is the property of Michle Delsol (France), copyright 2023 USA and
// international. For any questions concerning use of the software contact
// Michle Delsol at CPParduino@md-dsl.fr.
// 
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include "Globals.h"

//Scenario:
// A company gets created; employees are hired and some leave the company (voluntarily or fired).
// As employees come in and out, employee instances get allocated memory via an overloaded
// new and added to a linked list.
// A list of employees in each company is printed.
// The overloaded new creates a mem alloc item which is added to the linked list of mem allocs.
// The linked list of mem allocs is then printed to reveal the status and origin of
// each allocation.
// 
// This scenario uses several frameworks:
//    Memory Management Framework which handles memory allocations and releases.
//    LinkedList Framework to manage employees, companies and the mem allocs.
//    Operator overloading to overload new so that it could create mem allocs and append them
//    to the mem allocs list.
//    The Class and Function Names Framework's sue is illustrated via a few implementations.
// 
// Before getting into the nitty gritty of using the above, a small test program (TestMemMgt)
// reports initial memory, allocates memory, reports, frees memory and reports again.
// 
// The first order of business was creating the Employee and Company classes.Having them
// inherit from the LinkedList classes simplified programming tremendously.
// The second order of business was creating the memory alloc classes which also inherit 
// linked lits. Individual mem allocs are added to the mem allocs linked list from inside
// the overloaded new operator.
// 
// 


#ifdef USE_MEM_ALLOC_TRACKING
//#define _DEBUG_A_TEST
// Create the memory allocations tracking list
MemAllocTrackingList* memAllocTrackingList = new MemAllocTrackingList("Mem tracking list", 0, EClass::none);
ClassNames* classNames = new ClassNames();
void TestMemMgt(); // Function defined after ATest
#endif // USE_MEM_ALLOC_TRACKING

// ----------------------------------------  ATEST ----------------------------------------
class NewTest {
public:
	NewTest(uint16_t _dummyValue) : dummyValue(_dummyValue) {}
	void PrintDummyValue(){ PRINT_VALUE_EQUAL_NL(dummyValue) }
	uint16_t dummyValue;
};

void ATest() {

	PRINT_FUNCTION_ENTER_NL;

	//NewTest* newTest = new NewTest(11);
	//delete (newTest);
	//return;

//#define DO_PRELIMINARY_MEMORY_TESTS
#if defined(DO_PRELIMINARY_MEMORY_TESTS) && defined(USE_MEM_ALLOC_TRACKING)
	// TEST MEMORY MANAGEMENT
	TestMemMgt();

	// FreeRAM is total free space - fragmented and contiguous
	uint16_t freeRAM{ FreeRAM(__FUNCTION__, __LINE__) };
	uint16_t heapSpace{ HeapSpace(__FUNCTION__, __LINE__) };
#ifdef _DEBUG_A_TEST
	PRINT_VALUE_EQUAL_NL(freeRAM);
	PRINT_VALUE_EQUAL_NL(heapSpace);
	PRINT_VALUE_EQUAL_NL((freeRAM - heapSpace));
#endif // _DEBUG_A_TEST

	PRINT_FUNCTION_EXIT_NL;
#endif // DO_PRELIMINARY_MEMORY_TESTS


	//Create a company and a few employees 
#ifdef USE_MEM_ALLOC_TRACKING
	Company* bestHoney = (Company*) new(memAllocTrackingList->GetItemID(), CLASS_INDEX_USE(Company), EAllocStatus::allocated, EAllocType::Company, EAllocType::e_global) Company("The Best Honey, Ltd.");
	Company* perfectGizmo = new (memAllocTrackingList->GetItemID(), CLASS_INDEX_USE(Company), EAllocStatus::allocated, EAllocType::Company, EAllocType::e_global) Company("Perfect Gizmo, Inc.");
	Employee* michele = new (memAllocTrackingList->GetItemID(), CLASS_INDEX_USE(Employee), EAllocStatus::allocated, EAllocType::Employee, EAllocType::e_global)Employee("Michele", EDept::production, bestHoney->GetItemID());
	bestHoney->AddEmployee(michele);
	Employee* sasha = new (memAllocTrackingList->GetItemID(), CLASS_INDEX_USE(Employee), EAllocStatus::allocated, EAllocType::Employee, EAllocType::e_global)Employee("Sasha", EDept::electrical, perfectGizmo->GetItemID());
	perfectGizmo->AddEmployee(sasha);
	Employee* fred = new (memAllocTrackingList->GetItemID(), CLASS_INDEX_USE(Employee), EAllocStatus::allocated, EAllocType::Employee, EAllocType::e_global) Employee("Fred", EDept::dataProcessing, perfectGizmo->GetItemID());
	perfectGizmo->AddEmployee(fred);
	Employee* mary = new (memAllocTrackingList->GetItemID(), CLASS_INDEX_USE(Employee), EAllocStatus::allocated, EAllocType::Employee, EAllocType::e_global) Employee("Mary", EDept::administration, perfectGizmo->GetItemID());
	perfectGizmo->AddEmployee(mary);
	Employee* olivia = new (memAllocTrackingList->GetItemID(), CLASS_INDEX_USE(Employee), EAllocStatus::allocated, EAllocType::Employee, EAllocType::e_global) Employee("Olivia", EDept::research, perfectGizmo->GetItemID());
	perfectGizmo->AddEmployee(olivia);
#else
	Company* bestHoney = new Company("The Best Honey, Ltd.");
	Employee* michele = new Employee("Michele", EDept::production, bestHoney->GetItemID()); bestHoney->AddEmployee(michele);
	Company* perfectGizmo = new Company("Perfect Gizmo, Inc.");
	Employee* sasha = new Employee("Sasha", EDept::electrical, perfectGizmo->GetItemID()); perfectGizmo->AddEmployee(sasha);
	Employee* fred = new Employee("Fred", EDept::dataProcessing, perfectGizmo->GetItemID()); perfectGizmo->AddEmployee(fred);
	Employee* mary = new Employee("Mary", EDept::administration, perfectGizmo->GetItemID()); perfectGizmo->AddEmployee(mary);
	Employee* olivia = new Employee("Olivia", EDept::research, perfectGizmo->GetItemID()); perfectGizmo->AddEmployee(olivia);
#endif // USE_MEM_ALLOC_TRACKING
	PRINT_VALUE_EQUAL_DESC_NL("Personnel list of ", bestHoney->GetName());
	bestHoney->PrintList();
	PRINT_VALUE_EQUAL_DESC_NL("Personnel list of ", perfectGizmo->GetName());
	perfectGizmo->PrintList();

#ifdef USE_MEM_ALLOC_TRACKING
	memAllocTrackingList->PrintList();
#endif // USE_MEM_ALLOC_TRACKING
	// Delete Mary
	delete (mary);
#ifdef USE_MEM_ALLOC_TRACKING
	memAllocTrackingList->PrintList();
#endif // USE_MEM_ALLOC_TRACKING
	return;

} // ATest ()

#ifdef USE_MEM_ALLOC_TRACKING
void TestMemMgt() {
	// Exercises FreeRAM and HeapSpace to define fragmented and contiguous memory
// Calculate average size of alloc for 100 allocs to fill free ram
// Will randomly allocate

// FreeRAM calculates total heap spaces, fragmented and contiguous
	uint16_t freeRam = FreeRAM(__FUNCTION__, __LINE__);
	PRINT_TEXT_VALUE_EQUAL("Calculated initial RAM", freeRam);
	// HeapSpace - Calculates heap space by brute force - allocate small chunks of memory 'til run out.
	uint16_t heapSpace = HeapSpace(__FUNCTION__, __LINE__);
	PRINT_TEXT_VALUE_EQUAL("Brute force initial HEAP", heapSpace);

#define NUMBER_OF_ALLOCS 40
	uint16_t maxAlloc = (freeRam / NUMBER_OF_ALLOCS * 2) + 10;
	char* allocArray[NUMBER_OF_ALLOCS];

	// initialize array to 0
	for (uint8_t i = 0; i < NUMBER_OF_ALLOCS; i++) {
		allocArray[i] = 0;
	} // if

	// Allocate randomly
	Serial.print(F("Allocating randomly, no. allocs = ")); Serial.println(NUMBER_OF_ALLOCS);
	for (uint16_t i = 0; i < NUMBER_OF_ALLOCS; i++) {
		uint16_t sizeToAlloc = random(1, maxAlloc);
		char* addressAllocatedTo = (char*)malloc(random(1, maxAlloc));
		if (addressAllocatedTo == (char*)0) {
			Serial.print(F("i = ")); Serial.print(sizeToAlloc);
			Serial.print(F(" - Could not allocate ")); Serial.print(sizeToAlloc); Serial.println(F(" bytes."));
		}
		else {
			allocArray[i] = addressAllocatedTo;
			Serial.print(F("i = ")); Serial.print(i);
			Serial.print(F(" - sizeToAlloc = ")); Serial.print(sizeToAlloc);
			Serial.print(F(" - allocArray[i] = ")); Serial.println((uint16_t)allocArray[i]);
		}
	} // if
	freeRam = FreeRAM(__FUNCTION__, __LINE__);
	PRINT_TEXT_VALUE_EQUAL("Calculated RAM after allocs ", freeRam);
	heapSpace = HeapSpace(__FUNCTION__, __LINE__);
	PRINT_TEXT_VALUE_EQUAL("Brute force HEAP after allocs ", heapSpace);


	// free one of two
	for (uint16_t i = 0; i < NUMBER_OF_ALLOCS; i++) {
		if (allocArray[i] != 0) {
			free(allocArray[i]);
		}
		//    Serial.print(F("i = ")); Serial.print(i);
		//    Serial.print(F(" - sizeToAlloc = ")); Serial.print(sizeToAlloc);
		//    Serial.print(F(" - allocArray[i] = ")); Serial.println((uint16_t)allocArray[i]);
	} // if

	// FreeRAM calculates total heap spaces, fragmented and contiguous
	freeRam = FreeRAM(__FUNCTION__, __LINE__);
	PRINT_TEXT_VALUE_EQUAL("Calculated RAM after freeing", freeRam);
	// HeapSpace - Calculates heap space by brute force - allocate small chunks of memory 'til run out.
	heapSpace = HeapSpace(__FUNCTION__, __LINE__);
	PRINT_TEXT_VALUE_EQUAL("Brute force HEAP after freeing", heapSpace);

} // TestMemMgt

//*****************************************************
#define _DEBUG_A_TEST
#ifdef _DEBUG_A_TEST
#endif // _DEBUG_A_TEST

// ----------------------------------------  OPERATOR NEW ----------------------------------------

#endif // USE_MEM_ALLOC_TRACKING

// End ATest.cpp
