// Begin MemFunctions.cpp

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// incorporate it inside of an executable or other machine language component
// without restriction for personal or commercial use.The Software may not be
// redistributed as source code or in any recognizable human readable form
// in any form whatsoever for any use whatsoever.
// 
// The Software is the property of Michle Delsol (France), copyright 2023 USA and
// international. For any questions concerning use of the software contact
// Michle Delsol at CPParduino@md-dsl.fr.
// 
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include "Globals.h"

//************************************** SRAM *********************************************

extern uint16_t __data_start;
extern uint16_t __data_end;
extern uint16_t __bss_end;
extern uint16_t __heap_start;
extern void* __brkval;



uint16_t GetFreeRAM() {
#ifdef _DEBUG_MEM_MGT_GET_FREE_SRAM
	PRINT_FUNCTION_ENTER_NL
#endif
		uint8_t newVariable{ 0 };
	uint16_t freeRamFromStackPtr{ (uint16_t)&newVariable };
	// newVariable created to define stack pointer
	// heap is empty, use bss as start memory address
	if ((uint16_t)__brkval == 0) {
		freeRamFromStackPtr -= (uint16_t)__bss_end;
		// use heap end as the start of the memory address
	}
	else {
		freeRamFromStackPtr -= (uint16_t)__brkval;
	}
#ifdef _DEBUG_MEM_MGT_GET_FREE_SRAM
	Serial.print(F("&newVariable = "));  Serial.println((long)&newVariable);
	Serial.print(F("__brkval = "));  Serial.println((long)__brkval);
	//Serial.print (F ("&__brkval = "));  Serial.println ( (long)&__brkval);
	Serial.print(F("__bss_end = "));  Serial.println((long)__bss_end);
	Serial.print(F("freeRamFromStackPtr = "));  Serial.println(freeRamFromStackPtr);
	PRINT_FUNCTION_EXIT
#endif // _DEBUG_MEM_MGT_FREE_SRAM
		return freeRamFromStackPtr;
} // GetFreeRAM ()

uint16_t FreeRAM(char* _callingFunction, int _lineNo) {
	// Calculates total heap spaces, fragmented and contiguous
	// Fragmented memory calculated by traveling up FreeList (add hole sizes)
	// Contiguous memory calculated by
#ifdef _DEBUG_MEM_MGT_FREE_SRAM
	PRINT_FUNCTION_ENTER_NL
#endif
		uint16_t maxHole = 0;
	uint16_t fragmentedMemory{ 0 };
	uint16_t contiguousMemory{ 0 };
	contiguousMemory = (uint16_t)&contiguousMemory + sizeof(contiguousMemory);
	uint16_t totalHeapSpace{ 0 };
	// the last element in stack, so its address is
	// current pointer to end of stack

#if defined (__AVR__)
	extern uint16_t __heap_start;	// beginning of allocation heap
	extern void* __brkval;				// current end of heap (0 if no allocation yet)
	extern struct FreeList* __flp;	// head of the free list structure

	if (__brkval == 0) {
		contiguousMemory -= (uint16_t)&__heap_start;
	}
	else {
		contiguousMemory -= (uint16_t)__brkval;
		/* add all free memory fragments if any */
		for (struct FreeList* current = __flp; current; current = current->next) {
			fragmentedMemory += 2;		// two bytes for the memory block's header
			uint16_t holeSize = (uint16_t)current->size;
			maxHole = holeSize > maxHole ? holeSize : maxHole;
			fragmentedMemory += holeSize;
		}
	}
#endif // defined (__AVR__)
	totalHeapSpace = contiguousMemory + fragmentedMemory;
	uint16_t fragRatio = (uint16_t)((float)fragmentedMemory / (float)totalHeapSpace * (float)100);
#ifdef _DEBUG_MEM_MGT_FREE_SRAM
	PRINT_VALUE_EQUAL(maxHole);
	PRINT_VALUE_EQUAL(fragRatio);
	PRINT_VALUE_EQUAL(contiguousMemory);
	PRINT_VALUE_EQUAL(fragmentedMemory);
	PRINT_VALUE_EQUAL(totalHeapSpace);
	PRINT_FUNCTION_EXIT
#endif // _DEBUG_MEM_MGT_FREE_SRAM
		return (totalHeapSpace);
} // FreeRAM

// ----------------------------------------  SIMPLE HEAP SPACED ----------------------------------------
// These are needed when not using the MemMgt class

uint16_t HeapSpace(char* _callingFunction, int _lineNo) {
	// Calculates heap space by brute force - allocate small chunks of memory 'til run out.
	// allocates repeatedly memory via malloc to get lower and upper bounds
	// 'till they converge to a difference of 1 byte, malloc'ed mem is freed
	// function consumes 8 bytes of RAM to run
	// kick of is RAM size depending on chip
	// Overhead at start of setup of TestProgram on a Uno:
	//   no prints 8*1024 = 8192 - 7797 = 395
	//   compiler no prints reports 242 bytes
	//   delta is an extra 153 bytes consumed by program
#ifdef _DEBUG_MEM_MGT_HEAP_SPACE
	PRINT_FUNCTION_ENTER_NL;
#endif // _DEBUG_MEM_MGT_HEAP_SPACE

#ifdef ARDUINO_MEGA
	uint16_t maxSupposedMemory{ 8 * 1024 };
#endif
#ifdef ARDUINO_328
	uint16_t maxSupposedMemory{ 2 * 1024 };
#endif
	// If algorithm can't allocate 10 bytes it stops the process.
	// Do not want to count wasted memory under 10 bytes
	const uint16_t heapAllocationCutoff = 0; // Under 10 bytes considered not usable, could be fragmented
	uint16_t minPossibleMemAlloc{ 0 }; // heapsize when 10 bytes left
	int counter{ 0 };
	// heapSize is the ultimate available heapsize;
	///irrespective of fragmenting, will alloc all the way to 1 byte
	uint16_t heapSize{ 1 }; // Start process with a malloc of 1
	uint16_t deltaHeapSize{ maxSupposedMemory };
	int multiplier{ 1 };
	// Mallocs so that pointers be non zero after freeing

	//byte* heapPointer = (byte)malloc (1);
	byte* heapPointer = new byte[1]; // Init 1 to force multiplier to 1
	uint16_t heapSizePositive{ heapSize };
	uint16_t heapSizeNegative{ maxSupposedMemory };
	while (true) {
		//******************************************************
		counter++;
		multiplier = heapPointer == NULL ? -1 : 1; // if null pointer need decrease heap
#ifdef _DEBUG_MEM_MGT_HEAP_SPACE
		Serial.print(F("\theapSizeNegative ")); Serial.println(heapSizeNegative);
		Serial.print(F("\theapSizePositive ")); Serial.println(heapSizePositive);
#endif // _DEBUG_MEM_MGT_HEAP_SPACE
		if (heapSizePositive > heapSizeNegative) {
#ifdef _DEBUG_MEM_MGT_HEAP_SPACE
			Serial.println(F("ERROR: heapSizePositive greater than heapSizeNegative "));
#endif // _DEBUG_MEM_MGT_HEAP_SPACE
			return 0;
		}
		deltaHeapSize = (heapSizeNegative - heapSizePositive) / 2;
		heapSize = heapSize + deltaHeapSize * multiplier;
#ifdef _DEBUG_MEM_MGT_HEAP_SPACE
		Serial.print(F("\tdeltaHeapSize ")); Serial.println(deltaHeapSize);
		Serial.print(F("\theapSize ")); Serial.println(heapSize);
#endif // _DEBUG_MEM_MGT_HEAP_SPACE
		if (deltaHeapSize >= heapAllocationCutoff) {
			// This is what gets returned
#ifdef _DEBUG_MEM_MGT_HEAP_SPACE
			Serial.print(F("Redefining minPossibleMemAlloc to  ")); Serial.println(heapSize);
#endif
			minPossibleMemAlloc = heapSize;
		}
#ifdef _DEBUG_MEM_MGT_HEAP_SPACE
		Serial.print(F("counter ")); Serial.print(counter);
		Serial.print(F(", multiplier ")); Serial.print(multiplier);
		//Serial.print (F ("\theapPointer ")); Serial.println ( (uint16_t)heapPointer);
		Serial.print(F("\theapPointer ")); Serial.println((uint16_t)heapPointer);
		Serial.print(F("\tminPossibleMemAlloc ")); Serial.println((uint16_t)minPossibleMemAlloc);
#endif
		delete[] heapPointer;
		heapPointer = (byte*)malloc(heapSize);
		if (heapPointer == NULL) {
			// Could not allocate memory sized heapSize
			heapSizeNegative = heapSize;
		}
		else {
			// Successful allocation of memory sized heapSize
			//free (heapPointer); heapPointer = 0;
			heapSizePositive = heapSize;
		}
#ifdef _DEBUG_MEM_MGT_HEAP_SPACE
		Serial.print(F("\tJust did malloc (heapSize) ")); Serial.println((uint16_t)heapSize);
		Serial.print(F("\theapSizeNegative ")); Serial.print((uint16_t)heapSizeNegative);
		Serial.print(F("\theapSizePositive ")); Serial.print((uint16_t)heapSizePositive);
#endif
		//******************************************************/
		if (deltaHeapSize < 1) {
			delete[] heapPointer;
			heapPointer = 0;
			break;
		}
	} // While

#ifdef _DEBUG_MEM_MGT_HEAP_SPACE
  //Serial.print (F ("beginningOfHeap ")); Serial.print ( (uint16_t)beginningOfHeap);
  //Serial.print (F (", *endOfHeap ")); Serial.print ( (uint16_t)endOfHeap);
	Serial.print(F(", heapSizePositive ")); Serial.print(heapSizePositive);
	Serial.print(F(", minPossibleMemAlloc ")); Serial.println(minPossibleMemAlloc);
#endif
	//if (heapPointer != NULL) {
	//delay (100);
	//free (heapPointer);
	//}
#ifdef _DEBUG_MEM_MGT_HEAP_SPACE
	Serial.print(F("counter = ")); Serial.println(counter);
	Serial.print(F("returning with heapSizePositive = ")); Serial.println(heapSizePositive);
	PRINT_FUNCTION_EXIT
#endif
#ifdef _DEBUG_MEM_MGT_HEAP_SPACE
		PRINT_FUNCTION_EXIT_NL;
#endif // _DEBUG_MEM_MGT_HEAP_SPACE
	return heapSizePositive;
} // HeapSpace (_functionName)

void PrintHeapSpace(char* _description, char* _callingFunction, int _lineNo) {
	// Call this function to print the three heap space calculation methods
#ifdef _DEBUG_MEM_MGT_HEAP_SPACE
	PRINT_FUNCTION_ENTER_NL;
#endif // _DEBUG_MEM_MGT_HEAP_SPACE
	Serial.println(_description);
	uint16_t heapSpace = HeapSpace(__FUNCTION__, __LINE__);
	uint16_t getFreeSram = GetFreeRAM();
	uint16_t freeSram = FreeRAM(__FUNCTION__, __LINE__);
	Serial.print(F("heapSpace ")); Serial.println(heapSpace);
	Serial.print(F("getFreeSram ")); Serial.println(getFreeSram);
	Serial.print(F("freeSram ")); Serial.println(freeSram);
#ifdef _DEBUG_MEM_MGT_HEAP_SPACE
	PRINT_FUNCTION_EXIT_NL;
#endif // _DEBUG_MEM_MGT_HEAP_SPACE
} // PrintHeapSpace ()


// End MemFunctions.cpp
