#ifndef MEMORY_MGT_H
#define MEMORY_MGT_H

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// incorporate it inside of an executable or other machine language component
// without restriction for personal or commercial use.The Software may not be
// redistributed as source code or in any recognizable human readable form
// in any form whatsoever for any use whatsoever.
// 
// The Software is the property of Michle Delsol (France), copyright 2023 USA and
// international. For any questions concerning use of the software contact
// Michle Delsol at CPParduino@md-dsl.fr.
// 
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include "Globals.h"

#ifdef USE_MEM_ALLOC_TRACKING

// This memory management framework contains functions and classes to monitor memory use.
// It has three groups:
// 1 - Memory availability functions - determine heap space.
// 2 - Memory allocations tracking  - 
// 2 - Memory management classes - 
// 
// 
//   Support enums EnumMemAllocType and EnumMemCreateRemove for reporting
//   ClassName and ClassNameList - they serve to construct a list of class
//   names which will later be used for reporting
//   ObjectCreationFollowUp - a unidirectional linked list of creation/deletion data
//   Functions - HeapSpace and PrintHeap space serve to determine how much heap is left.

// -------------------------------------- MEMORY AVAILABILITY FUNCTIONS ---------------------------------------------------
// HeapSpace and PrintHeap space serve to determine how much heap is left. 
// It does not inform on absolute heap use. Its algorithm tries to allocate 
// memory from large to small up to a certain limit - default is 10.
// When it finally cannot allocate any more memory, it stops and returns 
// the result.
// 
// -------------------------------------- MEMORY ALLOCATION TRACKING ---------------------------------------------------
// MemAllocTracking Item and List
// As objects get created, they are assigned a unique ID to track their status.
//
// -------------------------------------- MEMORY MANAGEMENT ---------------------------------------------------
// The MemMgt class provides functions to monitor memory use. It gets 
// instantiated only once. As objects get allocated/deallocated onto the heap,
// the info is stored in an array to subsequently examine how the application
// used the heap. This monitoring feature is turned on via a macro. When th
// e application is deemed to be free of potential memory issues, turning
// the USE_MEM_MGT macro off frees memory from unnecessary functionalities.
//
// 

//--------------------------------  SUPPORT ENUMS ----------------------------------------------------

enum class EAllocStatus : uint8_t {
	start = 0,
	allocated = 0,
	deleted,
	end
};
extern char* eAllocStatusLabels[];

enum class EAllocType : uint8_t {
	start = 0,
	undefined = 0,
	e_global, // For context only
	e_uint16_t,
	e_char,
	Employee,
	Company,
	end
};
extern char* eAllocTypeLabels[];

//--------------------------------  MEM ALLOC TRACKING ----------------------------------------------------

class MemAllocTrackingItem : public LinkedListItem {
public:
	MemAllocTrackingItem(uint16_t _itemID, uint16_t* _itemAddress,
		EAllocStatus _allocStatus, EAllocType _allocType, EAllocType _allocContext);
	~MemAllocTrackingItem() {};
	uint16_t* GetItemAddress() { return itemAddress; }
	void PrintItem();
	void SetStatus(EAllocStatus _status) { allocStatus = _status; };
private:
	uint16_t* itemAddress;
	EAllocStatus allocStatus{ EAllocStatus::allocated };
	EAllocType allocType{ EAllocType::undefined };
	EAllocType allocContext{ EAllocType::undefined }; // Where in defined
}; // class MemAllocTrackingItem

class MemAllocTrackingList : public LinkedListManage {
public:
	MemAllocTrackingList(char* _listName, uint16_t _itemID, CLASS_INDEX_PARAMETER)
		: LinkedListManage(_listName, _itemID, CLASS_INDEX_PARAMETER_TRANSFER) {};
	void AddMemAlloc(MemAllocTrackingItem* _trackingAlloc);
	void DeleteMemAlloc(MemAllocTrackingItem* _trackingAlloc);
	void* FindItem(void* _itemPtr);
	void PrintList();
private:
	// These are the net number of objects active at any given time new and malloc
	// Each object creation gets a unique new number, these are never decremented
}; // class MemAllocTrackingList


//regex to find declaration: ^([ ]*\w+ *)( +|\*+|&)( *)([ ]*\w+)

#endif // USE_MEM_ALLOC_TRACKING

#endif // MEMORY_MGT_H
