/* ATest.cpp
   called by Aardvark to test stuff
*/

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// incorporate it inside of an executable or other machine language component
// without restriction for personal or commercial use.The Software may not be
// redistributed as source code or in any recognizable human readable form
// in any form whatsoever for any use whatsoever.
// 
// The Software is the property of Michle Delsol(France), copyright 2023 USA and
// international. For any questions concerning use of the software contact
// Michle Delsol at CPParduino@md-dsl.fr.
// 
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include "Globals.h"

//#define _DEBUG_ATEST_ATEST

// env for Primes setjmp/longjmp
jmp_buf primesJmp;


// Scenario
// The objective is to test the setjmp/longjmp mechanism and error reporting.
// I imagined a prime numbers generator which would test several conditions 
// to handle errors and report on them.
// My definition of a prime number is: an integer number divisible by 1 or by
// itself only. Based on this definition, the first prime numbers would be 1, 2,
// 3, 5, 7, 11, 13, 17, 19...
// To find whether a number is a prime number is easy: divide the number by all 
// previous numbers and determine whether there is a remainder. In programming terms, 
// this means do a modulo on the number with prior numbers ands check the result. 
// If any of the prior numbers yield zero on the modulo operation, the number 
// is not a prime because it is divisible by a prior number. 
// The problem with this brute force approach is processing time. There are two
// considerations to factor in to shorten processing time:
// 
// 1 - Limit the search to the square root of the number you are checking. The
// reason for this is that if a number is not a prime number, it is the product
// of two numbers. If you take the square root of the number, one of the two 
// numbers will be less than the other, hence a prime number which will have
// already been detected. For example, suppose you are checking
// whether 27 is a prime number. Take the square root of 27 - result is 5
// (6 yields 36) - 26 is the product of 2 and 13, 28 is the product of 7 
// and 2 used twice. QED, no need to check whether a number is divisible by 
// anything greater than its square root to determine whther it is a prime
// a number. This simple fact reduces the search through existing prime 
// numbers list.
// 
// 2 - As you progress finding prime numbers, check the number against prior 
// prime number. No need to check a number against a prior number which has 
// been found not to be a prime number. This implies appending a prime number
// just found to a list of existing prime numbers. One solution would be to 
// create an array of 1000 prime numbers or 10000 or more depending on memory 
// space available. But this renders the application not portable because you 
// would need to predefine the array's size according to memory size. A better
// solution would be to create a linked list of prime numbers which would 
// automatically be stopped when memory availability reached some threshold. 
// This is where the pseudo exception handling mechanism comes in. When the
// list gets so large that the computer cannot handle it any more, roll the 
// application back and do something. 
// 
// This application test scenario implements:
//   setjmp/longjmp jumps back to a landing point upon detecting some condition.
//   Since the program seems to be running well and is bug-free, it is not activated;
//   but you can artificially activate it and see how execution transfers down to the
//   bottom of the ATest()function (handleErro: label);
//   The GetAvailableSpace() memory function reports on available contiguous memory.
//   
// Caveats : Upon  starting the application, I calculate how many primes can be stored
// in an Arduino Uno. Result is noPrimesStorable = 274, yet upon running the program,
// I manage to generate 154 primes only. That is 120 primes short of theoretical storage.
// One explanation might be stack space requirements as the program runs which I might not 
// entirely account for. Getting a handle on why I encounter this problem requires that
// I implement hardware-based debugging which I have not done. 
//
//

uint8_t errorDetected{ 0 };

uint8_t memoryUseOverhead =
+ sizeof(EErrors)
+ sizeof(LinkedListItem)
+ sizeof(LinkedListManage)
+ 23 // primesErrorsLabels
+ 12 // primesList name
;

uint16_t GetAvailableSpace(){
		uint8_t * smallAlloc = malloc(1);
		long bottomMemory = (long)smallAlloc; 
		long topMemory = (long)&smallAlloc; // 
		long availableMemory = topMemory - bottomMemory + 3*sizeof(long) + 5; // freed malloc
		PRINT_VALUE_EQUAL_NL((long)bottomMemory);
		PRINT_VALUE_EQUAL_NL((long)topMemory);
		PRINT_VALUE_EQUAL_NL(availableMemory);
		free(smallAlloc);
		return availableMemory;
} // NAMELESS BLOCK END


void ATest() {
#ifdef _DEBUG_ATEST_ATEST
	PRINT_FUNCTION_ENTER_NL;
#endif // _DEBUG_ATEST_ATEST

	// Determine available memory and memory requirements
	uint16_t noPrimesStorable{ 0 };
	long availableMemory = GetAvailableSpace();
	noPrimesStorable = (availableMemory - memoryUseOverhead) / sizeof(LinkedListItem);
	uint16_t theoreticalMemory = (2048 - 328);
	long unaccountedMemory = theoreticalMemory - availableMemory;
	PRINT_VALUE_EQUAL_NL(noPrimesStorable);
#ifdef _DEBUG_ATEST_ATEST
	PRINT_VALUE_EQUAL_NL(memoryUseOverhead);
	PRINT_VALUE_EQUAL_NL(theoreticalMemory);
	PRINT_VALUE_EQUAL_DESC_NL("Bytes unaccounted for : ", unaccountedMemory);
#endif // _DEBUG_ATEST_ATEST

	// Manage prime numbers
	errorDetected = setjmp(primesJmp);
	if (errorDetected != 0) {
		// Error detected - hence go to error handling at end to function
		goto handleError;
	}
	PrimesList primesList = PrimesList(1000, "Primes list");
	primesList.GeneratePrimes();
	primesList.PrintPrimes();
#ifdef _DEBUG_ATEST_ATEST
	PRINT_FUNCTION_EXIT_NL;
#endif // _DEBUG_ATEST_ATEST
	return;
handleError:
	PRINT_VALUE_EQUAL_NL(errorDetected);
	PRINT_VALUE_EQUAL_NL(primesErrorsLabels[errorDetected]);
	if (errorDetected == (uint8_t)EErrors::allocFail) {
		PrimeItem* lastPrimeItem = primesList.GetTail()->GetPrevious();
		uint16_t lastPrime = lastPrimeItem->GetPrimeNumber();
		uint16_t targetNumber = primesList.GetLimit();
		uint16_t primesCount = primesList.GetItemCount();
		PRINT_VALUE_EQUAL_NL(primesCount);
		PRINT_VALUE_EQUAL_NL(targetNumber);
		PRINT_VALUE_EQUAL_NL(lastPrime);
	}
	availableMemory = GetAvailableSpace();
	PRINT_VALUE_EQUAL_NL(availableMemory);
	primesList.PrintPrimes();
	PRINT_FUNCTION_EXIT_NL;
	return;
} // ATest()

// End ATest.cpp
